# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# Copyright The NiPreps Developers <nipreps@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# We support and encourage derived works from this project, please read
# about our expectations at
#
#     https://www.nipreps.org/community/licensing/
#
import os
import typing as ty

import nibabel as nb
from nipype.interfaces import utility as niu
from nipype.pipeline import engine as pe
from niworkflows.func.util import init_enhance_and_skullstrip_bold_wf
from niworkflows.interfaces.header import ValidateImage
from niworkflows.interfaces.nitransforms import ConcatenateXFMs
from niworkflows.interfaces.utility import KeySelect
from sdcflows.workflows.apply.correction import init_unwarp_wf
from sdcflows.workflows.apply.registration import init_coeff2epi_wf

from nibabies import config
from nibabies._types import Anatomical
from nibabies.interfaces.reports import FunctionalSummary
from nibabies.interfaces.resampling import (
    DistortionParameters,
    ReconstructFieldmap,
    ResampleSeries,
)
from nibabies.utils.bids import extract_entities
from nibabies.utils.misc import estimate_bold_mem_usage

if ty.TYPE_CHECKING:
    from bids.layout import BIDSLayout

# BOLD workflows
from nibabies.workflows.bold.hmc import init_bold_hmc_wf
from nibabies.workflows.bold.outputs import (
    init_ds_boldref_wf,
    init_ds_hmc_wf,
    init_ds_registration_wf,
    init_func_fit_reports_wf,
)
from nibabies.workflows.bold.reference import init_raw_boldref_wf
from nibabies.workflows.bold.registration import init_bold_reg_wf
from nibabies.workflows.bold.stc import init_bold_stc_wf
from nibabies.workflows.bold.t2s import init_bold_t2s_wf


def get_sbrefs(
    bold_files: list[str],
    entity_overrides: dict[str, ty.Any],
    layout: 'BIDSLayout',
) -> list[str]:
    """Find single-band reference(s) associated with BOLD file(s)

    Parameters
    ----------
    bold_files
        List of absolute paths to BOLD files
    entity_overrides
        Query parameters to override defaults
    layout
        :class:`~bids.layout.BIDSLayout` to query

    Returns
    -------
    sbref_files
        List of absolute paths to sbref files associated with input BOLD files,
        sorted by EchoTime
    """
    entities = extract_entities(bold_files)
    entities.pop('echo', None)
    entities.update(suffix='sbref', extension=['.nii', '.nii.gz'])
    entities.update(entity_overrides)

    return sorted(
        layout.get(return_type='file', **entities),
        key=lambda fname: layout.get_metadata(fname).get('EchoTime'),
    )


def init_bold_fit_wf(
    *,
    bold_series: list[str],
    reference_anat: Anatomical,
    precomputed: dict | None = None,
    fieldmap_id: str | None = None,
    omp_nthreads: int = 1,
    name: str = 'bold_fit_wf',
) -> pe.Workflow:
    """
    This workflow controls the minimal estimation steps for functional preprocessing.

    Workflow Graph
        .. workflow::
            :graph2use: orig
            :simple_form: yes

            from nibabies.workflows.tests import mock_config
            from nibabies import config
            from nibabies.workflows.bold.fit import init_bold_fit_wf
            with mock_config():
                bold_file = config.execution.bids_dir / "sub-01" / "func" \
                    / "sub-01_task-mixedgamblestask_run-01_bold.nii.gz"
                wf = init_bold_fit_wf(bold_series=[str(bold_file)], reference_anat='T1w')

    Parameters
    ----------
    bold_series
        List of paths to NIfTI files, sorted by echo time.
    precomputed
        Dictionary containing precomputed derivatives to reuse, if possible.
    fieldmap_id
        ID of the fieldmap to use to correct this BOLD series. If :obj:`None`,
        no correction will be applied.

    Inputs
    ------
    bold_file
        BOLD series NIfTI file
    anat_preproc
        Bias-corrected structural template image
    anat_mask
        Mask of the skull-stripped template image
    anat_dseg
        Segmentation of preprocessed structural image, including
        gray-matter (GM), white-matter (WM) and cerebrospinal fluid (CSF)
    anat2std_xfm
        List of transform files, collated with templates
    subjects_dir
        FreeSurfer SUBJECTS_DIR
    subject_id
        FreeSurfer subject ID
    fsnative2t1w_xfm
        LTA-style affine matrix translating from FreeSurfer-conformed subject space to T1w
    fmap_id
        Unique identifiers to select fieldmap files
    fmap
        List of estimated fieldmaps (collated with fmap_id)
    fmap_ref
        List of fieldmap reference files (collated with fmap_id)
    fmap_coeff
        List of lists of spline coefficient files (collated with fmap_id)
    fmap_mask
        List of fieldmap masks (collated with fmap_id)
    sdc_method
        List of fieldmap correction method names (collated with fmap_id)

    Outputs
    -------
    hmc_boldref
        BOLD reference image used for head motion correction.
        Minimally processed to ensure consistent contrast with BOLD series.
    coreg_boldref
        BOLD reference image used for coregistration. Contrast-enhanced
        and fieldmap-corrected for greater anatomical fidelity, and aligned
        with ``hmc_boldref``.
    bold_mask
        Mask of ``coreg_boldref``.
    motion_xfm
        Affine transforms from each BOLD volume to ``hmc_boldref``, written
        as concatenated ITK affine transforms.
    boldref2anat_xfm
        Affine transform mapping from BOLD reference space to the anatomical
        space.
    boldref2fmap_xfm
        Affine transform mapping from BOLD reference space to the fieldmap
        space, if applicable.
    movpar_file
        MCFLIRT motion parameters, normalized to SPM format (X, Y, Z, Rx, Ry, Rz)
    rmsd_file
        Root mean squared deviation as measured by ``fsl_motion_outliers`` [Jenkinson2002]_.
    dummy_scans
        The number of dummy scans declared or detected at the beginning of the series.

    See Also
    --------

    * :py:func:`~nibabies.workflows.bold.reference.init_raw_boldref_wf`
    * :py:func:`~nibabies.workflows.bold.hmc.init_bold_hmc_wf`
    * :py:func:`~niworkflows.func.utils.init_enhance_and_skullstrip_bold_wf`
    * :py:func:`~sdcflows.workflows.apply.registration.init_coeff2epi_wf`
    * :py:func:`~sdcflows.workflows.apply.correction.init_unwarp_wf`
    * :py:func:`~nibabies.workflows.bold.registration.init_bold_reg_wf`
    * :py:func:`~nibabies.workflows.bold.outputs.init_ds_boldref_wf`
    * :py:func:`~nibabies.workflows.bold.outputs.init_ds_hmc_wf`
    * :py:func:`~nibabies.workflows.bold.outputs.init_ds_registration_wf`

    """
    from niworkflows.engine.workflows import LiterateWorkflow as Workflow

    from nibabies.utils.misc import estimate_bold_mem_usage

    if precomputed is None:
        precomputed = {}
    layout = config.execution.layout
    bids_filters = config.execution.get().get('bids_filters', {})

    # Fitting operates on the shortest echo
    # This could become more complicated in the future
    bold_file = bold_series[0]

    # Collect sbref files, sorted by EchoTime
    sbref_files = get_sbrefs(
        bold_series,
        entity_overrides=bids_filters.get('sbref', {}),
        layout=layout,
    )

    basename = os.path.basename(bold_file)
    sbref_msg = f'No single-band-reference found for {basename}.'
    if sbref_files and 'sbref' in config.workflow.ignore:
        sbref_msg = f'Single-band reference file(s) found for {basename} and ignored.'
        sbref_files = []
    elif sbref_files:
        sbref_msg = 'Using single-band reference file(s) {}.'.format(
            ','.join([os.path.basename(sbf) for sbf in sbref_files])
        )
    config.loggers.workflow.info(sbref_msg)

    # Get metadata from BOLD file(s)
    entities = extract_entities(bold_series)
    metadata = layout.get_metadata(bold_file)
    orientation = ''.join(nb.aff2axcodes(nb.load(bold_file).affine))

    bold_tlen, mem_gb = estimate_bold_mem_usage(bold_file)

    # Boolean used to update workflow self-descriptions
    multiecho = len(bold_series) > 1

    have_hmcref = 'hmc_boldref' in precomputed
    have_coregref = 'coreg_boldref' in precomputed
    # Can contain
    #  1) boldref2fmap
    #  2) boldref2anat
    #  3) hmc
    transforms = precomputed.get('transforms', {})
    hmc_xforms = transforms.get('hmc')
    boldref2fmap_xform = transforms.get('boldref2fmap')
    boldref2anat_xform = transforms.get('boldref2anat')

    workflow = Workflow(name=name)

    inputnode = pe.Node(
        niu.IdentityInterface(
            fields=[
                'bold_file',
                # Fieldmap registration
                'fmap',
                'fmap_ref',
                'fmap_coeff',
                'fmap_mask',
                'fmap_id',
                'sdc_method',
                # Anatomical coregistration
                'anat_preproc',
                'anat_mask',
                'anat_dseg',
                'subjects_dir',
                'subject_id',
                'fsnative2anat_xfm',
            ],
        ),
        name='inputnode',
    )
    inputnode.inputs.bold_file = bold_series

    outputnode = pe.Node(
        niu.IdentityInterface(
            fields=[
                'dummy_scans',
                'hmc_boldref',
                'coreg_boldref',
                'bold_mask',
                'motion_xfm',
                'boldref2anat_xfm',
                'boldref2fmap_xfm',
                'movpar_file',
                'rmsd_file',
            ],
        ),
        name='outputnode',
    )

    # If all derivatives exist, inputnode could go unconnected, so add explicitly
    workflow.add_nodes([inputnode])

    hmcref_buffer = pe.Node(
        niu.IdentityInterface(fields=['boldref', 'bold_file', 'dummy_scans']),
        name='hmcref_buffer',
    )
    fmapref_buffer = pe.Node(niu.Function(function=_select_ref), name='fmapref_buffer')
    hmc_buffer = pe.Node(
        niu.IdentityInterface(fields=['hmc_xforms', 'movpar_file', 'rmsd_file']), name='hmc_buffer'
    )
    fmapreg_buffer = pe.Node(
        niu.IdentityInterface(fields=['boldref2fmap_xfm']), name='fmapreg_buffer'
    )
    regref_buffer = pe.Node(
        niu.IdentityInterface(fields=['boldref', 'boldmask']), name='regref_buffer'
    )

    use_fs = config.workflow.surface_recon_method is not None
    summary = pe.Node(
        FunctionalSummary(
            distortion_correction='None',  # Can override with connection
            registration='FreeSurfer' if use_fs else 'FSL',
            registration_dof=config.workflow.bold2anat_dof,
            registration_init=config.workflow.bold2anat_init,
            pe_direction=metadata.get('PhaseEncodingDirection'),
            echo_idx=entities.get('echo', []),
            tr=metadata['RepetitionTime'],
            orientation=orientation,
        ),
        name='summary',
        mem_gb=config.DEFAULT_MEMORY_MIN_GB,
        run_without_submitting=True,
    )
    summary.inputs.dummy_scans = config.workflow.dummy_scans
    if config.workflow.level == 'full':
        # Hack. More pain than it's worth to connect this up at a higher level.
        # We can consider separating out fit and transform summaries,
        # or connect a bunch a bunch of summary parameters to outputnodes
        # to make available to the base workflow.
        summary.inputs.slice_timing = (
            bool(metadata.get('SliceTiming')) and 'slicetiming' not in config.workflow.ignore
        )

    func_fit_reports_wf = init_func_fit_reports_wf(
        reference_anat=reference_anat,
        # TODO: Enable sdc report even if we find coregref
        sdc_correction=not (have_coregref or fieldmap_id is None),
        output_dir=config.execution.output_dir,
    )

    workflow.connect([
        (hmcref_buffer, outputnode, [
            ('boldref', 'hmc_boldref'),
            ('dummy_scans', 'dummy_scans'),
        ]),
        (regref_buffer, outputnode, [
            ('boldref', 'coreg_boldref'),
            ('boldmask', 'bold_mask'),
        ]),
        (fmapreg_buffer, outputnode, [('boldref2fmap_xfm', 'boldref2fmap_xfm')]),
        (hmc_buffer, outputnode, [
            ('hmc_xforms', 'motion_xfm'),
            ('movpar_file', 'movpar_file'),
            ('rmsd_file', 'rmsd_file'),
        ]),
        (inputnode, func_fit_reports_wf, [
            ('bold_file', 'inputnode.source_file'),
            ('anat_preproc', 'inputnode.anat_preproc'),
            # May not need all of these
            ('anat_mask', 'inputnode.anat_mask'),
            ('anat_dseg', 'inputnode.anat_dseg'),
            ('subjects_dir', 'inputnode.subjects_dir'),
            ('subject_id', 'inputnode.subject_id'),
        ]),
        (outputnode, func_fit_reports_wf, [
            ('coreg_boldref', 'inputnode.coreg_boldref'),
            ('bold_mask', 'inputnode.bold_mask'),
            ('boldref2anat_xfm', 'inputnode.boldref2anat_xfm'),
        ]),
        (summary, func_fit_reports_wf, [('out_report', 'inputnode.summary_report')]),
    ])  # fmt:skip

    # Stage 1: Generate motion correction boldref
    hmc_boldref_source_buffer = pe.Node(
        niu.IdentityInterface(fields=['in_file']),
        name='hmc_boldref_source_buffer',
    )
    if not have_hmcref:
        config.loggers.workflow.info('Stage 1: Adding HMC boldref workflow')
        hmc_boldref_wf = init_raw_boldref_wf(
            name='hmc_boldref_wf',
            bold_file=bold_file,
            multiecho=multiecho,
            ref_frame_start=config.workflow.hmc_bold_frame,
        )
        hmc_boldref_wf.inputs.inputnode.dummy_scans = config.workflow.dummy_scans

        ds_hmc_boldref_wf = init_ds_boldref_wf(
            output_dir=config.execution.output_dir,
            desc='hmc',
            name='ds_hmc_boldref_wf',
        )
        ds_hmc_boldref_wf.inputs.inputnode.source_files = [bold_file]

        workflow.connect([
            (hmc_boldref_wf, hmcref_buffer, [
                ('outputnode.bold_file', 'bold_file'),
                ('outputnode.boldref', 'boldref'),
                ('outputnode.skip_vols', 'dummy_scans'),
            ]),
            (hmcref_buffer, ds_hmc_boldref_wf, [('boldref', 'inputnode.boldref')]),
            (hmc_boldref_wf, summary, [('outputnode.algo_dummy_scans', 'algo_dummy_scans')]),
            (hmc_boldref_wf, func_fit_reports_wf, [
                ('outputnode.validation_report', 'inputnode.validation_report'),
            ]),
            (ds_hmc_boldref_wf, hmc_boldref_source_buffer, [
                ('outputnode.boldref', 'in_file'),
            ]),
        ])  # fmt:skip
    else:
        config.loggers.workflow.info('Found HMC boldref - skipping Stage 1')

        validate_bold = pe.Node(ValidateImage(), name='validate_bold')
        validate_bold.inputs.in_file = bold_file

        hmcref_buffer.inputs.boldref = precomputed['hmc_boldref']

        workflow.connect([
            (validate_bold, hmcref_buffer, [('out_file', 'bold_file')]),
            (validate_bold, func_fit_reports_wf, [('out_report', 'inputnode.validation_report')]),
            (hmcref_buffer, hmc_boldref_source_buffer, [('boldref', 'in_file')]),
        ])  # fmt:skip

    # Stage 2: Estimate head motion
    if not hmc_xforms:
        config.loggers.workflow.info('Stage 2: Adding motion correction workflow')
        bold_hmc_wf = init_bold_hmc_wf(
            name='bold_hmc_wf', mem_gb=mem_gb['filesize'], omp_nthreads=omp_nthreads
        )

        ds_hmc_wf = init_ds_hmc_wf(output_dir=config.execution.output_dir)
        ds_hmc_wf.inputs.inputnode.source_files = [bold_file]

        workflow.connect([
            (hmcref_buffer, bold_hmc_wf, [
                ('boldref', 'inputnode.raw_ref_image'),
                ('bold_file', 'inputnode.bold_file'),
            ]),
            (bold_hmc_wf, ds_hmc_wf, [('outputnode.xforms', 'inputnode.xforms')]),
            (bold_hmc_wf, hmc_buffer, [
                ('outputnode.movpar_file', 'movpar_file'),
                ('outputnode.rmsd_file', 'rmsd_file'),
            ]),
            (ds_hmc_wf, hmc_buffer, [('outputnode.xforms', 'hmc_xforms')]),
        ])  # fmt:skip
    else:
        config.loggers.workflow.info('Found motion correction transforms - skipping Stage 2')
        hmc_buffer.inputs.hmc_xforms = hmc_xforms

    # Stage 3: Create coregistration reference
    # Fieldmap correction only happens during fit if this stage is needed
    if not have_coregref:
        config.loggers.workflow.info('Stage 3: Adding coregistration boldref workflow')

        # Select initial boldref, enhance contrast, and generate mask
        fmapref_buffer.inputs.sbref_files = sbref_files
        enhance_boldref_wf = init_enhance_and_skullstrip_bold_wf(omp_nthreads=omp_nthreads)

        ds_coreg_boldref_wf = init_ds_boldref_wf(
            output_dir=config.execution.output_dir,
            desc='coreg',
            name='ds_coreg_boldref_wf',
        )

        workflow.connect([
            (hmcref_buffer, fmapref_buffer, [('boldref', 'boldref_files')]),
            (fmapref_buffer, enhance_boldref_wf, [('out', 'inputnode.in_file')]),
            (hmc_boldref_source_buffer, ds_coreg_boldref_wf, [
                ('in_file', 'inputnode.source_files'),
            ]),
            (ds_coreg_boldref_wf, regref_buffer, [('outputnode.boldref', 'boldref')]),
            (fmapref_buffer, func_fit_reports_wf, [('out', 'inputnode.sdc_boldref')]),
        ])  # fmt:skip

        if fieldmap_id:
            fmap_select = pe.Node(
                KeySelect(
                    fields=['fmap_ref', 'fmap_coeff', 'fmap_mask', 'sdc_method'],
                    key=fieldmap_id,
                ),
                name='fmap_select',
                run_without_submitting=True,
            )

            if not boldref2fmap_xform:
                fmapreg_wf = init_coeff2epi_wf(
                    debug='fieldmaps' in config.execution.debug,
                    omp_nthreads=config.nipype.omp_nthreads,
                    sloppy=config.execution.sloppy,
                    name='fmapreg_wf',
                )

                itk_mat2txt = pe.Node(ConcatenateXFMs(out_fmt='itk'), name='itk_mat2txt')

                ds_fmapreg_wf = init_ds_registration_wf(
                    output_dir=config.execution.output_dir,
                    source='boldref',
                    dest=fieldmap_id.replace('_', ''),
                    name='ds_fmapreg_wf',
                )
                ds_fmapreg_wf.inputs.inputnode.source_files = [bold_file]

                workflow.connect([
                    (enhance_boldref_wf, fmapreg_wf, [
                        ('outputnode.bias_corrected_file', 'inputnode.target_ref'),
                        ('outputnode.mask_file', 'inputnode.target_mask'),
                    ]),
                    (fmap_select, fmapreg_wf, [
                        ('fmap_ref', 'inputnode.fmap_ref'),
                        ('fmap_mask', 'inputnode.fmap_mask'),
                    ]),
                    (fmapreg_wf, itk_mat2txt, [('outputnode.target2fmap_xfm', 'in_xfms')]),
                    (itk_mat2txt, ds_fmapreg_wf, [('out_xfm', 'inputnode.xform')]),
                    (ds_fmapreg_wf, fmapreg_buffer, [('outputnode.xform', 'boldref2fmap_xfm')]),
                ])  # fmt:skip
            else:
                fmapreg_buffer.inputs.boldref2fmap_xfm = boldref2fmap_xform

            unwarp_wf = init_unwarp_wf(
                free_mem=config.environment.free_mem,
                debug='fieldmaps' in config.execution.debug,
                omp_nthreads=config.nipype.omp_nthreads,
            )
            unwarp_wf.inputs.inputnode.metadata = layout.get_metadata(bold_file)

            workflow.connect([
                (inputnode, fmap_select, [
                    ('fmap_ref', 'fmap_ref'),
                    ('fmap_coeff', 'fmap_coeff'),
                    ('fmap_mask', 'fmap_mask'),
                    ('sdc_method', 'sdc_method'),
                    ('fmap_id', 'keys'),
                ]),
                (fmap_select, unwarp_wf, [
                    ('fmap_coeff', 'inputnode.fmap_coeff'),
                ]),
                (fmapreg_buffer, unwarp_wf, [
                    # This looks backwards, but unwarp_wf describes transforms in
                    # terms of points while we (and init_coeff2epi_wf) describe them
                    # in terms of images. Mapping fieldmap coordinates into boldref
                    # coordinates maps the boldref image onto the fieldmap image.
                    ('boldref2fmap_xfm', 'inputnode.fmap2data_xfm'),
                ]),
                (enhance_boldref_wf, unwarp_wf, [
                    ('outputnode.bias_corrected_file', 'inputnode.distorted'),
                ]),
                (unwarp_wf, ds_coreg_boldref_wf, [
                    ('outputnode.corrected', 'inputnode.boldref'),
                ]),
                (unwarp_wf, regref_buffer, [
                    ('outputnode.corrected_mask', 'boldmask'),
                ]),
                (fmap_select, func_fit_reports_wf, [('fmap_ref', 'inputnode.fmap_ref')]),
                (fmap_select, summary, [('sdc_method', 'distortion_correction')]),
                (fmapreg_buffer, func_fit_reports_wf, [
                    ('boldref2fmap_xfm', 'inputnode.boldref2fmap_xfm'),
                ]),
                (unwarp_wf, func_fit_reports_wf, [('outputnode.fieldmap', 'inputnode.fieldmap')]),
            ])  # fmt:skip
        else:
            workflow.connect([
                (enhance_boldref_wf, ds_coreg_boldref_wf, [
                    ('outputnode.bias_corrected_file', 'inputnode.boldref'),
                ]),
                (enhance_boldref_wf, regref_buffer, [
                    ('outputnode.mask_file', 'boldmask'),
                ]),
            ])  # fmt:skip
    else:
        config.loggers.workflow.info('Found coregistration reference - skipping Stage 3')
        regref_buffer.inputs.boldref = precomputed['coreg_boldref']

    if not boldref2anat_xform:
        # calculate BOLD registration to anat
        bold_reg_wf = init_bold_reg_wf(
            bold2anat_dof=config.workflow.bold2anat_dof,
            bold2anat_init=config.workflow.bold2anat_init,
            use_bbr=config.workflow.use_bbr,
            freesurfer=use_fs,
            omp_nthreads=omp_nthreads,
            mem_gb=mem_gb['resampled'],
            sloppy=config.execution.sloppy,
        )

        ds_boldreg_wf = init_ds_registration_wf(
            output_dir=config.execution.output_dir,
            source='boldref',
            dest=reference_anat,
            name='ds_boldreg_wf',
        )

        workflow.connect([
            (inputnode, bold_reg_wf, [
                ('anat_preproc', 'inputnode.anat_preproc'),
                ('anat_mask', 'inputnode.anat_mask'),
                ('anat_dseg', 'inputnode.anat_dseg'),
                # Undefined if --fs-no-reconall, but this is safe
                ('subjects_dir', 'inputnode.subjects_dir'),
                ('subject_id', 'inputnode.subject_id'),
                ('fsnative2anat_xfm', 'inputnode.fsnative2anat_xfm'),
            ]),
            (regref_buffer, bold_reg_wf, [('boldref', 'inputnode.ref_bold_brain')]),
            # Incomplete sources
            (regref_buffer, ds_boldreg_wf, [('boldref', 'inputnode.source_files')]),
            (bold_reg_wf, ds_boldreg_wf, [('outputnode.itk_bold_to_anat', 'inputnode.xform')]),
            (ds_boldreg_wf, outputnode, [('outputnode.xform', 'boldref2anat_xfm')]),
            (bold_reg_wf, summary, [('outputnode.fallback', 'fallback')]),
        ])  # fmt:skip
    else:
        outputnode.inputs.boldref2anat_xfm = boldref2anat_xform

    return workflow


def init_bold_native_wf(
    *,
    bold_series: list[str],
    fieldmap_id: str | None = None,
    omp_nthreads: int = 1,
    name: str = 'bold_native_wf',
) -> pe.Workflow:
    r"""
    Minimal resampling workflow.

    This workflow performs slice-timing correction, and resamples to boldref space
    with head motion and susceptibility distortion correction. It also handles
    multi-echo processing and selects the transforms needed to perform further
    resampling.

    Workflow Graph
        .. workflow::
            :graph2use: orig
            :simple_form: yes

            from nibabies.workflows.tests import mock_config
            from nibabies import config
            from nibabies.workflows.bold.fit import init_bold_native_wf
            with mock_config():
                bold_file = config.execution.bids_dir / "sub-01" / "func" \
                    / "sub-01_task-mixedgamblestask_run-01_bold.nii.gz"
                wf = init_bold_native_wf(bold_series=[str(bold_file)])

    Parameters
    ----------
    bold_series
        List of paths to NIfTI files.
    fieldmap_id
        ID of the fieldmap to use to correct this BOLD series. If :obj:`None`,
        no correction will be applied.

    Inputs
    ------
    boldref
        BOLD reference file
    bold_mask
        Mask of BOLD reference file
    motion_xfm
        Affine transforms from each BOLD volume to ``hmc_boldref``, written
        as concatenated ITK affine transforms.
    boldref2fmap_xfm
        Affine transform mapping from BOLD reference space to the fieldmap
        space, if applicable.
    fmap_id
        Unique identifiers to select fieldmap files
    fmap_ref
        List of fieldmap reference files (collated with fmap_id)
    fmap_coeff
        List of lists of spline coefficient files (collated with fmap_id)

    Outputs
    -------
    bold_minimal
        BOLD series ready for further resampling. For single-echo data, only
        slice-timing correction (STC) may have been applied. For multi-echo
        data, this is identical to bold_native.
    bold_native
        BOLD series resampled into BOLD reference space. Slice-timing,
        head motion and susceptibility distortion correction (STC, HMC, SDC)
        will all be applied to each file. For multi-echo data, the echos
        are combined to form an `optimal combination`_.
    metadata
        Metadata dictionary of BOLD series with the shortest echo
    motion_xfm
        Motion correction transforms for further correcting bold_minimal.
        For multi-echo data, motion correction has already been applied, so
        this will be undefined.
    bold_echos
        The individual, corrected echos, suitable for use in Tedana.
        (Multi-echo only.)
    t2star_map
        The T2\* map estimated by Tedana when calculating the optimal combination.
        (Multi-echo only.)

    See Also
    --------

    * :py:func:`~nibabies.workflows.bold.stc.init_bold_stc_wf`
    * :py:func:`~nibabies.workflows.bold.t2s.init_bold_t2s_wf`

    .. _optimal combination: https://tedana.readthedocs.io/en/stable/approach.html#optimal-combination

    """

    layout = config.execution.layout

    # Shortest echo first
    all_metadata = [layout.get_metadata(bold_file) for bold_file in bold_series]
    echo_times = [md.get('EchoTime') for md in all_metadata]
    multiecho = len(bold_series) > 1

    bold_file = bold_series[0]
    metadata = all_metadata[0]

    _, mem_gb = estimate_bold_mem_usage(bold_file)

    if multiecho:
        shapes = [nb.load(echo).shape for echo in bold_series]
        if len(set(shapes)) != 1:
            diagnostic = '\n'.join(
                f'{os.path.basename(echo)}: {shape}'
                for echo, shape in zip(bold_series, shapes, strict=False)
            )
            raise RuntimeError(f'Multi-echo images found with mismatching shapes\n{diagnostic}')
        if len(shapes) == 2:
            raise RuntimeError(
                'Multi-echo processing requires at least three different echos (found two).'
            )

    run_stc = bool(metadata.get('SliceTiming')) and 'slicetiming' not in config.workflow.ignore

    workflow = pe.Workflow(name=name)

    inputnode = pe.Node(
        niu.IdentityInterface(
            fields=[
                # BOLD fit
                'boldref',
                'bold_mask',
                'motion_xfm',
                'boldref2fmap_xfm',
                'dummy_scans',
                # Fieldmap fit
                'fmap_ref',
                'fmap_coeff',
                'fmap_id',
            ],
        ),
        name='inputnode',
    )

    outputnode = pe.Node(
        niu.IdentityInterface(
            fields=[
                'bold_minimal',
                'bold_native',
                'metadata',
                # Transforms
                'motion_xfm',
                # Multiecho outputs
                'bold_echos',  # Individual corrected echos
                't2star_map',  # T2* map
            ],  # fmt:skip
        ),
        name='outputnode',
    )
    outputnode.inputs.metadata = metadata

    boldbuffer = pe.Node(
        niu.IdentityInterface(fields=['bold_file', 'ro_time', 'pe_dir']), name='boldbuffer'
    )

    # Track echo index - this allows us to treat multi- and single-echo workflows
    # almost identically
    echo_index = pe.Node(niu.IdentityInterface(fields=['echoidx']), name='echo_index')
    if multiecho:
        echo_index.iterables = [('echoidx', range(len(bold_series)))]
    else:
        echo_index.inputs.echoidx = 0

    # BOLD source: track original BOLD file(s)
    bold_source = pe.Node(niu.Select(inlist=bold_series), name='bold_source')
    validate_bold = pe.Node(ValidateImage(), name='validate_bold')
    workflow.connect([
        (echo_index, bold_source, [('echoidx', 'index')]),
        (bold_source, validate_bold, [('out', 'in_file')]),
    ])  # fmt:skip

    # Slice-timing correction
    if run_stc:
        bold_stc_wf = init_bold_stc_wf(metadata=metadata)
        workflow.connect([
            (inputnode, bold_stc_wf, [('dummy_scans', 'inputnode.skip_vols')]),
            (validate_bold, bold_stc_wf, [('out_file', 'inputnode.bold_file')]),
            (bold_stc_wf, boldbuffer, [('outputnode.stc_file', 'bold_file')]),
        ])  # fmt:skip
    else:
        workflow.connect([(validate_bold, boldbuffer, [('out_file', 'bold_file')])])

    # Prepare fieldmap metadata
    if fieldmap_id:
        fmap_select = pe.Node(
            KeySelect(fields=['fmap_ref', 'fmap_coeff'], key=fieldmap_id),
            name='fmap_select',
            run_without_submitting=True,
        )

        distortion_params = pe.Node(
            DistortionParameters(metadata=metadata, in_file=bold_file),
            name='distortion_params',
            run_without_submitting=True,
        )
        workflow.connect([
            (inputnode, fmap_select, [
                ('fmap_ref', 'fmap_ref'),
                ('fmap_coeff', 'fmap_coeff'),
                ('fmap_id', 'keys'),
            ]),
            (distortion_params, boldbuffer, [
                ('readout_time', 'ro_time'),
                ('pe_direction', 'pe_dir'),
            ]),
        ])  # fmt:skip

    # Resample to boldref
    boldref_bold = pe.Node(
        ResampleSeries(
            jacobian='fmap-jacobian' not in config.workflow.ignore,
            num_threads=omp_nthreads,
        ),
        name='boldref_bold',
        n_procs=omp_nthreads,
        mem_gb=mem_gb['resampled'],
    )

    workflow.connect([
        (inputnode, boldref_bold, [
            ('boldref', 'ref_file'),
            ('motion_xfm', 'transforms'),
        ]),
        (boldbuffer, boldref_bold, [
            ('bold_file', 'in_file'),
            ('ro_time', 'ro_time'),
            ('pe_dir', 'pe_dir'),
        ]),
    ])  # fmt:skip

    if fieldmap_id:
        boldref_fmap = pe.Node(ReconstructFieldmap(inverse=[True]), name='boldref_fmap', mem_gb=1)
        workflow.connect([
            (inputnode, boldref_fmap, [
                ('boldref', 'target_ref_file'),
                ('boldref2fmap_xfm', 'transforms'),
            ]),
            (fmap_select, boldref_fmap, [
                ('fmap_coeff', 'in_coeffs'),
                ('fmap_ref', 'fmap_ref_file'),
            ]),
            (boldref_fmap, boldref_bold, [('out_file', 'fieldmap')]),
        ])  # fmt:skip

    if multiecho:
        join_echos = pe.JoinNode(
            niu.IdentityInterface(fields=['bold_files']),
            joinsource='echo_index',
            joinfield=['bold_files'],
            name='join_echos',
            run_without_submitting=True,
        )

        # create optimal combination, adaptive T2* map
        bold_t2s_wf = init_bold_t2s_wf(
            echo_times=echo_times,
            mem_gb=mem_gb['filesize'],
            omp_nthreads=config.nipype.omp_nthreads,
            me_t2s_fit_method=config.workflow.me_t2s_fit_method,
            name='bold_t2smap_wf',
        )

        # Do NOT set motion_xfm on outputnode
        # This prevents downstream resamplers from double-dipping
        workflow.connect([
            (inputnode, bold_t2s_wf, [('bold_mask', 'inputnode.bold_mask')]),
            (boldref_bold, join_echos, [('out_file', 'bold_files')]),
            (join_echos, bold_t2s_wf, [('bold_files', 'inputnode.bold_file')]),
            (join_echos, outputnode, [('bold_files', 'bold_echos')]),
            (bold_t2s_wf, outputnode, [
                ('outputnode.bold', 'bold_minimal'),
                ('outputnode.bold', 'bold_native'),
                ('outputnode.t2star_map', 't2star_map'),
            ]),
        ])  # fmt:skip
    else:
        workflow.connect([
            (inputnode, outputnode, [('motion_xfm', 'motion_xfm')]),
            (boldbuffer, outputnode, [('bold_file', 'bold_minimal')]),
            (boldref_bold, outputnode, [('out_file', 'bold_native')]),
        ])  # fmt:skip

    return workflow


def _select_ref(sbref_files, boldref_files):
    """Select first sbref or boldref file, preferring sbref if available"""
    from niworkflows.utils.connections import listify

    refs = sbref_files or boldref_files
    return listify(refs)[0]



"""
The workflow builder factory method.

All the checks and the construction of the workflow are done
inside this function that has pickleable inputs and output
dictionary (``retval``) to allow isolation using a
``multiprocessing.Process`` that allows nibabies to enforce
a hard-limited memory-scope.

"""


def build_workflow(config_file):
    """Create the Nipype Workflow that supports the whole execution graph."""
    from niworkflows.utils.bids import check_pipeline_version
    from niworkflows.utils.misc import check_valid_fs_license

    from .. import config
    from ..reports.core import generate_reports
    from ..utils.misc import check_deps
    from ..workflows.base import init_nibabies_wf

    # initialize config
    config.load(config_file)
    build_logger = config.loggers.workflow

    nibabies_dir = config.execution.nibabies_dir
    version = config.environment.version

    retval = {'return_code': 1, 'workflow': None}

    # warn if older results exist: check for dataset_description.json in output folder
    msg = check_pipeline_version(version, nibabies_dir / 'dataset_description.json')
    if msg is not None:
        build_logger.warning(msg)

    # Please note this is the input folder's dataset_description.json
    dset_desc_path = config.execution.bids_dir / 'dataset_description.json'
    if dset_desc_path.exists():
        from hashlib import sha256

        desc_content = dset_desc_path.read_bytes()
        config.execution.bids_description_hash = sha256(desc_content).hexdigest()

    # Called with reports only
    if config.execution.reports_only:
        build_logger.log(
            25,
            'Running --reports-only on participants %s',
            ', '.join(config.execution.unique_labels),
        )
        retval['return_code'] = generate_reports(
            config.execution.unique_labels,
            nibabies_dir,
            config.execution.run_uuid,
        )
        return retval

    # Build main workflow
    init_msg = f"""
    Running nibabies version {config.environment.version}:
      * BIDS dataset path: {config.execution.bids_dir}.
      * Participant list: {config.execution.unique_labels}.
      * Run identifier: {config.execution.run_uuid}.
      * Output spaces: {config.execution.output_spaces or 'MNIInfant'}."""

    if config.execution.derivatives:
        init_msg += f"""
      * Derivatives: {config.execution.derivatives}."""

    if config.execution.fs_subjects_dir:
        init_msg += f"""
      * Pre-run FreeSurfer's SUBJECTS_DIR: {config.execution.fs_subjects_dir}."""
    build_logger.log(25, init_msg)

    retval['workflow'] = init_nibabies_wf(config.execution.unique_labels)

    # Check for FS license after building the workflow
    if not check_valid_fs_license():
        build_logger.critical(
            """\
ERROR: a valid license file is required for FreeSurfer to run. nibabies looked for an existing \
license file at several paths, in this order: 1) command line argument ``--fs-license-file``; \
2) ``$FS_LICENSE`` environment variable; and 3) the ``$FREESURFER_HOME/license.txt`` path. Get it \
(for free) by registering at https://surfer.nmr.mgh.harvard.edu/registration.html"""
        )
        retval['return_code'] = 126  # 126 == Command invoked cannot execute.
        return retval

    # Check workflow for missing commands
    missing = check_deps(retval['workflow'])
    if missing:
        build_logger.critical(
            'Cannot run nibabies. Missing dependencies:%s',
            '\n\t* '.join([''] + [f'{cmd} (Interface: {iface})' for iface, cmd in missing]),
        )
        retval['return_code'] = 127  # 127 == command not found.
        return retval

    # config.to_filename(config_file)
    build_logger.info(
        'NiBabies workflow graph with %d nodes built successfully.',
        len(retval['workflow']._get_all_nodes()),
    )
    retval['return_code'] = 0
    return retval


def build_boilerplate(workflow):
    """Write boilerplate in an isolated process."""
    from .. import config

    logs_path = config.execution.nibabies_dir / 'logs'
    boilerplate = workflow.visit_desc()
    citation_files = {ext: logs_path / f'CITATION.{ext}' for ext in ('bib', 'tex', 'md', 'html')}

    if boilerplate:
        # To please git-annex users and also to guarantee consistency
        # among different renderings of the same file, first remove any
        # existing one
        for citation_file in citation_files.values():
            try:
                citation_file.unlink()
            except FileNotFoundError:
                pass

    citation_files['md'].write_text(boilerplate)

    if not config.execution.md_only_boilerplate and citation_files['md'].exists():
        from shutil import copyfile
        from subprocess import CalledProcessError, TimeoutExpired, check_call

        from nibabies.data import load as load_data

        bib = str(load_data('boilerplate.bib'))

        # Generate HTML file resolving citations
        cmd = [
            'pandoc',
            '-s',
            '--bibliography',
            bib,
            '--citeproc',
            '--metadata',
            'pagetitle="nibabies citation boilerplate"',
            str(citation_files['md']),
            '-o',
            str(citation_files['html']),
        ]

        config.loggers.cli.info('Generating an HTML version of the citation boilerplate...')
        try:
            check_call(cmd, timeout=10)
        except (FileNotFoundError, CalledProcessError, TimeoutExpired):
            config.loggers.cli.warning('Could not generate CITATION.html file:\n%s', ' '.join(cmd))

        # Generate LaTex file resolving citations
        cmd = [
            'pandoc',
            '-s',
            '--bibliography',
            bib,
            '--natbib',
            str(citation_files['md']),
            '-o',
            str(citation_files['tex']),
        ]
        config.loggers.cli.info('Generating a LaTeX version of the citation boilerplate...')
        try:
            check_call(cmd, timeout=10)
        except (FileNotFoundError, CalledProcessError, TimeoutExpired):
            config.loggers.cli.warning('Could not generate CITATION.tex file:\n%s', ' '.join(cmd))
        else:
            copyfile(bib, citation_files['bib'])



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""Writing outputs."""

from __future__ import annotations

import typing as ty
from pathlib import Path

from nipype.interfaces import utility as niu
from nipype.pipeline import engine as pe
from niworkflows.engine import Workflow, tag
from niworkflows.interfaces.fixes import FixHeaderApplyTransforms as ApplyTransforms
from smriprep.workflows.outputs import init_template_iterator_wf

from nibabies.interfaces import DerivativesDataSink

if ty.TYPE_CHECKING:
    from niworkflows.utils.spaces import SpatialReferences

BIDS_TISSUE_ORDER = ('GM', 'WM', 'CSF')


@tag('anat.coreg-report')
def init_coreg_report_wf(*, output_dir, name='coreg_report_wf'):
    """
    Generate and store a report in the right location.

    Parameters
    ----------
    output_dir : :obj:`str`
        Directory in which to save derivatives
    name : :obj:`str`
        Workflow name (default: coreg_report_wf)

    Inputs
    ------
    source_file
        Input reference T1w image
    t1w_preproc
        Preprocessed T1w image.
    t2w_preproc
        Preprocessed T2w image, aligned with the T1w image.
    in_mask
        Brain mask.

    """
    from niworkflows.interfaces.reportlets.registration import (
        SimpleBeforeAfterRPT as SimpleBeforeAfter,
    )

    workflow = Workflow(name=name)

    inputfields = [
        'source_file',
        't1w_preproc',
        't2w_preproc',
        'in_mask',
    ]
    inputnode = pe.Node(niu.IdentityInterface(fields=inputfields), name='inputnode')
    # Generate reportlets showing spatial normalization
    norm_rpt = pe.Node(
        SimpleBeforeAfter(before_label='T2w', after_label='T1w'),
        name='norm_rpt',
        mem_gb=0.1,
    )

    ds_t1w_t2w_report = pe.Node(
        DerivativesDataSink(
            base_directory=output_dir, space='T2w', suffix='T1w', datatype='figures'
        ),
        name='ds_t1w_t2w_report',
        run_without_submitting=True,
    )

    # fmt: off
    workflow.connect([
        (inputnode, norm_rpt, [('t2w_preproc', 'before'),
                               ('t1w_preproc', 'after'),
                               ('in_mask', 'wm_seg')]),
        (inputnode, ds_t1w_t2w_report, [('source_file', 'source_file')]),
        (norm_rpt, ds_t1w_t2w_report, [('out_report', 'in_file')]),
    ])
    # fmt: on

    return workflow


@tag('anat.reports')
def init_anat_reports_wf(
    *,
    spaces: SpatialReferences,
    surface_recon: ty.Literal['freesurfer', 'infantfs', 'mcribs'] | None,
    output_dir: str,
    sloppy: bool,
    name='anat_reports_wf',
) -> Workflow:
    """
    Patched workflow for reports to allow no resolution for templates
    Set up a battery of datasinks to store reports in the right location.
    Parameters
    ----------
    recon_method : :obj:`bool`
        FreeSurfer was enabled
    output_dir : :obj:`str`
        Directory in which to save derivatives
    name : :obj:`str`
        Workflow name (default: anat_reports_wf)
    Inputs
    ------
    source_file
        Input T1w image
    std_t1w
        T1w image resampled to standard space
    std_mask
        Mask of skull-stripped template
    subject_dir
        FreeSurfer SUBJECTS_DIR
    subject_id
        FreeSurfer subject ID
    t1w_conform_report
        Conformation report
    t1w_preproc
        The T1w reference map, which is calculated as the average of bias-corrected
        and preprocessed T1w images, defining the anatomical space.
    t1w_dseg
        Segmentation in T1w space
    t1w_mask
        Brain (binary) mask estimated by brain extraction.
    template
        Template space and specifications
    """
    from niworkflows.interfaces.reportlets.masks import ROIsPlot
    from niworkflows.interfaces.reportlets.registration import (
        SimpleBeforeAfterRPT as SimpleBeforeAfter,
    )
    from smriprep.workflows.outputs import (
        _empty_report,
        _rpt_masks,
    )

    workflow = Workflow(name=name)

    inputfields = [
        'source_file',
        'anat_preproc',
        'anat_dseg',
        'anat_mask',
        'template',
        'anat2std_xfm',
        # Optional
        'subject_id',
        'subjects_dir',
        'anat_conform_report',
    ]
    inputnode = pe.Node(niu.IdentityInterface(fields=inputfields), name='inputnode')

    seg_rpt = pe.Node(ROIsPlot(colors=['b', 'magenta'], levels=[1.5, 2.5]), name='seg_rpt')

    anat_conform_check = pe.Node(
        niu.Function(function=_empty_report),
        name='anat_conform_check',
        run_without_submitting=True,
    )

    ds_anat_conform_report = pe.Node(
        DerivativesDataSink(base_directory=output_dir, desc='conform', datatype='figures'),
        name='ds_anat_conform_report',
        run_without_submitting=True,
    )

    ds_anat_dseg_mask_report = pe.Node(
        DerivativesDataSink(base_directory=output_dir, suffix='dseg', datatype='figures'),
        name='ds_anat_dseg_mask_report',
        run_without_submitting=True,
    )

    workflow.connect([
        (inputnode, anat_conform_check, [('anat_conform_report', 'in_file')]),
        (anat_conform_check, ds_anat_conform_report, [('out', 'in_file')]),
        (inputnode, ds_anat_conform_report, [('source_file', 'source_file')]),
        (inputnode, ds_anat_dseg_mask_report, [('source_file', 'source_file')]),
        (inputnode, seg_rpt, [('anat_preproc', 'in_file'),
                              ('anat_mask', 'in_mask'),
                              ('anat_dseg', 'in_rois')]),
        (seg_rpt, ds_anat_dseg_mask_report, [('out_report', 'in_file')]),
    ])  # fmt:skip

    if spaces._cached is not None and spaces.cached.references:
        template_iterator_wf = init_template_iterator_wf(spaces=spaces, sloppy=sloppy)
        t1w_std = pe.Node(
            ApplyTransforms(
                dimension=3,
                default_value=0,
                float=True,
                interpolation='LanczosWindowedSinc',
            ),
            name='t1w_std',
        )
        mask_std = pe.Node(
            ApplyTransforms(
                dimension=3,
                default_value=0,
                float=True,
                interpolation='MultiLabel',
            ),
            name='mask_std',
        )

        norm_msk = pe.Node(
            niu.Function(
                function=_rpt_masks,
                output_names=['before', 'after'],
                input_names=['mask_file', 'before', 'after', 'after_mask'],
            ),
            name='norm_msk',
        )
        norm_rpt = pe.Node(SimpleBeforeAfter(), name='norm_rpt', mem_gb=0.1)
        norm_rpt.inputs.after_label = 'Participant'  # after

        ds_std_t1w_report = pe.Node(
            DerivativesDataSink(base_directory=output_dir, suffix='T1w', datatype='figures'),
            name='ds_std_t1w_report',
            run_without_submitting=True,
        )

        workflow.connect([
            (inputnode, template_iterator_wf, [
                ('template', 'inputnode.template'),
                ('anat2std_xfm', 'inputnode.anat2std_xfm'),
            ]),
            (inputnode, t1w_std, [('anat_preproc', 'input_image')]),
            (inputnode, mask_std, [('anat_mask', 'input_image')]),
            (template_iterator_wf, t1w_std, [
                ('outputnode.anat2std_xfm', 'transforms'),
                ('outputnode.std_t1w', 'reference_image'),
            ]),
            (template_iterator_wf, mask_std, [
                ('outputnode.anat2std_xfm', 'transforms'),
                ('outputnode.std_t1w', 'reference_image'),
            ]),
            (template_iterator_wf, norm_rpt, [('outputnode.space', 'before_label')]),
            (t1w_std, norm_msk, [('output_image', 'after')]),
            (mask_std, norm_msk, [('output_image', 'after_mask')]),
            (template_iterator_wf, norm_msk, [
                ('outputnode.std_t1w', 'before'),
                ('outputnode.std_mask', 'mask_file'),
            ]),
            (norm_msk, norm_rpt, [
                ('before', 'before'),
                ('after', 'after'),
            ]),
            (inputnode, ds_std_t1w_report, [('source_file', 'source_file')]),
            (template_iterator_wf, ds_std_t1w_report, [('outputnode.space', 'space')]),
            (norm_rpt, ds_std_t1w_report, [('out_report', 'in_file')]),
        ])  # fmt:skip

    if not surface_recon:
        return workflow

    from smriprep.interfaces.reports import FSSurfaceReport

    recon_report = pe.Node(FSSurfaceReport(), name='recon_report')
    recon_report.interface._always_run = True

    if surface_recon == 'freesurfer':
        recon_desc = 'reconall'
    elif surface_recon == 'infantfs':
        recon_desc = 'infantfs'
    elif surface_recon == 'mcribs':
        recon_desc = 'mcribs'

    ds_recon_report = pe.Node(
        DerivativesDataSink(base_directory=output_dir, desc=recon_desc, datatype='figures'),
        name='ds_recon_report',
        run_without_submitting=True,
    )
    workflow.connect([
        (inputnode, recon_report, [('subjects_dir', 'subjects_dir'),
                                   ('subject_id', 'subject_id')]),
        (recon_report, ds_recon_report, [('out_report', 'in_file')]),
        (inputnode, ds_recon_report, [('source_file', 'source_file')])
    ])  # fmt: skip

    return workflow


@tag('anat.derivatives')
def init_anat_derivatives_wf(
    *,
    bids_root: Path | str,
    output_dir: Path | str,
    spaces: SpatialReferences,
    cifti_output: bool,
    num_t1w: int | None,
    num_t2w: int | None,
    surface_recon: ty.Literal['freesurfer', 'infantfs', 'mcribs'] | None,
    tpm_labels: tuple[str, str, str] = BIDS_TISSUE_ORDER,
    name: str = 'anat_derivatives_wf',
):
    """
    Set up a battery of datasinks to store derivatives in the right location.
    Parameters
    ----------
    bids_root : :obj:`str`
        Root path of BIDS dataset
    freesurfer : :obj:`bool`
        FreeSurfer was enabled
    num_t1w : :obj:`int`
        Number of T1w images
    output_dir : :obj:`str`
        Directory in which to save derivatives
    name : :obj:`str`
        Workflow name (default: anat_derivatives_wf)
    tpm_labels : :obj:`tuple`
        Tissue probability maps in order
    Inputs
    ------
    template
        Template space and specifications
    source_files
        List of input T1w images
    t1w_ref_xfms
        List of affine transforms to realign input T1w images
    t1w_preproc
        The T1w reference map, which is calculated as the average of bias-corrected
        and preprocessed T1w images, defining the anatomical space.
    t1w_mask
        Mask of the ``t1w_preproc``
    t1w_dseg
        Segmentation in T1w space
    t1w_tpms
        Tissue probability maps in T1w space
    anat2std_xfm
        Nonlinear spatial transform to resample imaging data given in anatomical space
        into standard space.
    std2anat_xfm
        Inverse transform of ``anat2std_xfm``
    std_t1w
        T1w reference resampled in one or more standard spaces.
    std_mask
        Mask of skull-stripped template, in standard space
    std_dseg
        Segmentation, resampled into standard space
    std_tpms
        Tissue probability maps in standard space
    t1w2fsnative_xfm
        LTA-style affine matrix translating from T1w to
        FreeSurfer-conformed subject space
    fsnative2t1w_xfm
        LTA-style affine matrix translating from FreeSurfer-conformed
        subject space to T1w
    surfaces
        GIFTI surfaces (gray/white boundary, midthickness, pial, inflated)
    t1w_fs_aseg
        FreeSurfer's aseg segmentation, in native T1w space
    t1w_fs_aparc
        FreeSurfer's aparc+aseg segmentation, in native T1w space
    t2w_source_files
        List of input T2w images
    t2w_preproc
        The T2w image in T1w space.
    cifti_morph
        Morphometric CIFTI-2 dscalar files
    cifti_density
        Grayordinate density
    cifti_metadata
        JSON files containing metadata dictionaries
    """
    from niworkflows.interfaces.nibabel import ApplyMask
    from niworkflows.interfaces.utility import KeySelect
    from smriprep.workflows.outputs import (
        _bids_relative,
        _combine_cohort,
        _drop_path,
        _fmt_cohort,
        _is_native,
    )

    workflow = Workflow(name=name)

    inputnode = pe.Node(
        niu.IdentityInterface(
            fields=[
                'template',
                # T1w
                't1w_source_files',
                't1w_ref_xfms',
                't1w_preproc',
                # T2w
                't2w_source_files',
                't2w_ref_xfms',
                't2w_preproc',
                # Can be in either T1w/T2w space
                'anat_mask',
                'anat_dseg',
                'anat_tpms',
                'anat2std_xfm',
                'std2anat_xfm',
                # FS
                'anat2fsnative_xfm',
                'fsnative2anat_xfm',
                'anat_fs_aseg',
                'anat_fs_aparc',
                'anat_ribbon',
                'surfaces',
                'morphometrics',
                # CIFTI
                'cifti_metadata',
                'cifti_density',
                'cifti_morph',
                'sphere_reg',
                'sphere_reg_fsLR',
            ]
        ),
        name='inputnode',
    )
    # The preferred space to use for to/from entities
    source_files = 't1w_source_files' if num_t1w else 't2w_source_files'
    space = 'T1w' if num_t1w else 'T2w'

    if num_t1w:
        raw_sources = pe.Node(niu.Function(function=_bids_relative), name='t1w_raw_sources')
        raw_sources.inputs.bids_root = bids_root

        ds_t1w_preproc = pe.Node(
            DerivativesDataSink(base_directory=output_dir, desc='preproc', compress=True),
            name='ds_t1w_preproc',
            run_without_submitting=True,
        )
        ds_t1w_preproc.inputs.SkullStripped = False

        if num_t1w > 1:
            # Please note the dictionary unpacking to provide the from argument.
            # It is necessary because from is a protected keyword (not allowed as argument name).
            ds_t1w_ref_xfms = pe.MapNode(
                DerivativesDataSink(
                    base_directory=output_dir,
                    to='T1w',
                    mode='image',
                    suffix='xfm',
                    extension='txt',
                    **{'from': 'orig'},
                ),
                iterfield=['source_file', 'in_file'],
                name='ds_t1w_ref_xfms',
                run_without_submitting=True,
            )
            # fmt:off
            workflow.connect([
                (inputnode, ds_t1w_ref_xfms, [('t1w_source_files', 'source_file'),
                                              ('t1w_ref_xfms', 'in_file')]),
            ])
            # fmt:on

    if num_t2w:
        if not num_t1w:
            raw_sources = pe.Node(niu.Function(function=_bids_relative), name='t2w_raw_sources')
            raw_sources.inputs.bids_root = bids_root

        ds_t2w_preproc = pe.Node(
            DerivativesDataSink(base_directory=output_dir, desc='preproc', compress=True),
            name='ds_t2w_preproc',
            run_without_submitting=True,
        )
        ds_t2w_preproc.inputs.SkullStripped = False

        if num_t1w:
            ds_t2w_preproc.inputs.space = 'T1w'

        if num_t2w > 1:
            # Please note the dictionary unpacking to provide the from argument.
            # It is necessary because from is a protected keyword (not allowed as argument name).
            ds_t2w_ref_xfms = pe.MapNode(
                DerivativesDataSink(
                    base_directory=output_dir,
                    to='T1w',
                    mode='image',
                    suffix='xfm',
                    extension='txt',
                    **{'from': 'orig'},
                ),
                iterfield=['source_file', 'in_file'],
                name='ds_t2w_ref_xfms',
                run_without_submitting=True,
            )
            # fmt:off
            workflow.connect([
                (inputnode, ds_t2w_ref_xfms, [('t2w_source_files', 'source_file'),
                                              ('t2w_ref_xfms', 'in_file')]),
            ])
            # fmt:on

    ds_anat_mask = pe.Node(
        DerivativesDataSink(base_directory=output_dir, desc='brain', suffix='mask', compress=True),
        name='ds_anat_mask',
        run_without_submitting=True,
    )
    ds_anat_mask.inputs.Type = 'Brain'

    ds_anat_dseg = pe.Node(
        DerivativesDataSink(base_directory=output_dir, suffix='dseg', compress=True),
        name='ds_anat_dseg',
        run_without_submitting=True,
    )

    ds_anat_tpms = pe.Node(
        DerivativesDataSink(base_directory=output_dir, suffix='probseg', compress=True),
        name='ds_anat_tpms',
        run_without_submitting=True,
    )
    ds_anat_tpms.inputs.label = tpm_labels

    ds_anat_ribbon = pe.Node(
        DerivativesDataSink(
            base_directory=output_dir,
            desc='ribbon',
            suffix='mask',
            extension='.nii.gz',
            compress=True,
        ),
        name='ds_anat_ribbon',
        run_without_submitting=True,
    )

    if num_t1w:
        workflow.connect([
            (inputnode, ds_t1w_preproc, [('t1w_preproc', 'in_file'),
                                         ('t1w_source_files', 'source_file')]),
        ])  # fmt:skip

    if num_t2w:
        workflow.connect([
            (inputnode, ds_t2w_preproc, [('t2w_preproc', 'in_file'),
                                         ('t2w_source_files', 'source_file')]),
        ])  # fmt:skip

    # fmt:off
    workflow.connect([
        (inputnode, raw_sources, [(source_files, 'in_files')]),
        (inputnode, ds_anat_mask, [('anat_mask', 'in_file'),
                                   (source_files, 'source_file')]),
        (inputnode, ds_anat_tpms, [('anat_tpms', 'in_file'),
                                   (source_files, 'source_file')]),
        (inputnode, ds_anat_dseg, [('anat_dseg', 'in_file'),
                                   (source_files, 'source_file')]),
        (inputnode, ds_anat_ribbon, [('anat_ribbon', 'in_file'),
                                     (source_files, 'source_file')]),
        (raw_sources, ds_anat_mask, [('out', 'RawSources')]),
    ])
    # fmt:on

    # Transforms
    if spaces.get_spaces(nonstandard=False, dim=(3,)):
        ds_std2anat_xfm = pe.MapNode(
            DerivativesDataSink(base_directory=output_dir, to=space, mode='image', suffix='xfm'),
            iterfield=('in_file', 'from'),
            name='ds_std2anat_xfm',
            run_without_submitting=True,
        )

        ds_anat2std_xfm = pe.MapNode(
            DerivativesDataSink(
                base_directory=output_dir, mode='image', suffix='xfm', **{'from': space}
            ),
            iterfield=('in_file', 'to'),
            name='ds_anat2std_xfm',
            run_without_submitting=True,
        )

        # fmt:off
        workflow.connect([
            (inputnode, ds_anat2std_xfm, [
                ('anat2std_xfm', 'in_file'),
                (('template', _combine_cohort), 'to'),
                (source_files, 'source_file')]),
            (inputnode, ds_std2anat_xfm, [
                ('std2anat_xfm', 'in_file'),
                (('template', _combine_cohort), 'from'),
                (source_files, 'source_file')]),
        ])
        # fmt:on

    # Write derivatives in standard spaces specified by --output-spaces
    if spaces._cached is not None and spaces.cached.references:
        from niworkflows.interfaces.fixes import (
            FixHeaderApplyTransforms as ApplyTransforms,
        )
        from niworkflows.interfaces.nibabel import GenerateSamplingReference
        from niworkflows.interfaces.space import SpaceDataSource
        from smriprep.interfaces.templateflow import TemplateFlowSelect

        spacesource = pe.Node(SpaceDataSource(), name='spacesource', run_without_submitting=True)
        spacesource.iterables = (
            'in_tuple',
            [(s.fullname, s.spec) for s in spaces.cached.get_standard(dim=(3,))],
        )

        gen_tplid = pe.Node(
            niu.Function(function=_fmt_cohort),
            name='gen_tplid',
            run_without_submitting=True,
        )

        select_xfm = pe.Node(
            KeySelect(fields=['anat2std_xfm']),
            name='select_xfm',
            run_without_submitting=True,
        )
        select_tpl = pe.Node(TemplateFlowSelect(), name='select_tpl', run_without_submitting=True)

        gen_ref = pe.Node(GenerateSamplingReference(), name='gen_ref', mem_gb=0.01)

        mask_anat = pe.Node(ApplyMask(), name='mask_anat')

        # Resample T1w-space inputs
        anat2std_t1w = pe.Node(
            ApplyTransforms(
                dimension=3,
                default_value=0,
                float=True,
                interpolation='LanczosWindowedSinc',
            ),
            name='anat2std_t1w',
        )

        anat2std_mask = pe.Node(ApplyTransforms(interpolation='MultiLabel'), name='anat2std_mask')
        anat2std_dseg = pe.Node(ApplyTransforms(interpolation='MultiLabel'), name='anat2std_dseg')
        anat2std_tpms = pe.MapNode(
            ApplyTransforms(dimension=3, default_value=0, float=True, interpolation='Gaussian'),
            iterfield=['input_image'],
            name='anat2std_tpms',
        )

        ds_std_t1w = pe.Node(
            DerivativesDataSink(
                base_directory=output_dir,
                desc='preproc',
                compress=True,
            ),
            name='ds_std_t1w',
            run_without_submitting=True,
        )
        ds_std_t1w.inputs.SkullStripped = True

        ds_std_mask = pe.Node(
            DerivativesDataSink(
                base_directory=output_dir, desc='brain', suffix='mask', compress=True
            ),
            name='ds_std_mask',
            run_without_submitting=True,
        )
        ds_std_mask.inputs.Type = 'Brain'

        ds_std_dseg = pe.Node(
            DerivativesDataSink(base_directory=output_dir, suffix='dseg', compress=True),
            name='ds_std_dseg',
            run_without_submitting=True,
        )

        ds_std_tpms = pe.Node(
            DerivativesDataSink(base_directory=output_dir, suffix='probseg', compress=True),
            name='ds_std_tpms',
            run_without_submitting=True,
        )

        set_tpl_res = pe.Node(
            niu.Function(function=_set_tpl_res),
            name='set_tpl_res',
            run_without_submitting=True,
            mem_gb=0.1,
        )

        # CRITICAL: the sequence of labels here (CSF-GM-WM) is that of the output of FSL-FAST
        #           (intensity mean, per tissue). This order HAS to be matched also by the ``tpms``
        #           output in the data/io_spec.json file.
        ds_std_tpms.inputs.label = tpm_labels

        preproc_file = 't1w_preproc' if num_t1w else 't2w_preproc'

        # fmt: off
        workflow.connect([
            (inputnode, mask_anat, [(preproc_file, 'in_file'),
                                    ('anat_mask', 'in_mask')]),
            (mask_anat, anat2std_t1w, [('out_file', 'input_image')]),
            (inputnode, anat2std_mask, [('anat_mask', 'input_image')]),
            (inputnode, anat2std_dseg, [('anat_dseg', 'input_image')]),
            (inputnode, anat2std_tpms, [('anat_tpms', 'input_image')]),
            (inputnode, gen_ref, [(preproc_file, 'moving_image')]),
            (inputnode, select_xfm, [
                ('anat2std_xfm', 'anat2std_xfm'),
                ('template', 'keys')]),
            (spacesource, gen_tplid, [('space', 'template'),
                                      ('cohort', 'cohort')]),
            (gen_tplid, select_xfm, [('out', 'key')]),
            (spacesource, select_tpl, [('space', 'template'),
                                       ('cohort', 'cohort')]),
            (spacesource, set_tpl_res, [('space', 'space'),
                                        ('resolution', 'resolution')]),
            (set_tpl_res, select_tpl, [('out', 'resolution')]),
            (spacesource, gen_ref, [(('resolution', _is_native), 'keep_native')]),
            (select_tpl, gen_ref, [('t1w_file', 'fixed_image')]),
            (anat2std_t1w, ds_std_t1w, [('output_image', 'in_file')]),
            (anat2std_mask, ds_std_mask, [('output_image', 'in_file')]),
            (anat2std_dseg, ds_std_dseg, [('output_image', 'in_file')]),
            (anat2std_tpms, ds_std_tpms, [('output_image', 'in_file')]),
            (select_tpl, ds_std_mask, [(('brain_mask', _drop_path), 'RawSources')]),
        ])
        # fmt: on
        workflow.connect(
            # Connect apply transforms nodes
            [
                (gen_ref, n, [('out_file', 'reference_image')])
                for n in (anat2std_t1w, anat2std_mask, anat2std_dseg, anat2std_tpms)
            ]
            + [
                (select_xfm, n, [('anat2std_xfm', 'transforms')])
                for n in (anat2std_t1w, anat2std_mask, anat2std_dseg, anat2std_tpms)
            ]
            # Connect the source_file input of these datasinks
            + [
                (inputnode, n, [(source_files, 'source_file')])
                for n in (ds_std_t1w, ds_std_mask, ds_std_dseg, ds_std_tpms)
            ]
            # Connect the space input of these datasinks
            + [
                (
                    spacesource,
                    n,
                    [('space', 'space'), ('cohort', 'cohort'), ('resolution', 'resolution')],
                )
                for n in (ds_std_t1w, ds_std_mask, ds_std_dseg, ds_std_tpms)
            ]
        )

    if not surface_recon:
        return workflow

    from niworkflows.interfaces.nitransforms import ConcatenateXFMs
    from niworkflows.interfaces.surf import Path2BIDS

    # FS native space transforms
    lta2itk_fwd = pe.Node(ConcatenateXFMs(), name='lta2itk_fwd', run_without_submitting=True)
    lta2itk_inv = pe.Node(ConcatenateXFMs(), name='lta2itk_inv', run_without_submitting=True)
    ds_anat_fsnative = pe.Node(
        DerivativesDataSink(
            base_directory=output_dir,
            mode='image',
            to='fsnative',
            suffix='xfm',
            extension='txt',
            **{'from': space},
        ),
        name='ds_anat_fsnative',
        run_without_submitting=True,
    )
    ds_fsnative_anat = pe.Node(
        DerivativesDataSink(
            base_directory=output_dir,
            mode='image',
            to=space,
            suffix='xfm',
            extension='txt',
            **{'from': 'fsnative'},
        ),
        name='ds_fsnative_anat',
        run_without_submitting=True,
    )
    # Surfaces
    name_surfs = pe.MapNode(
        Path2BIDS(), iterfield='in_file', name='name_surfs', run_without_submitting=True
    )
    ds_surfs = pe.MapNode(
        DerivativesDataSink(base_directory=output_dir, extension='.surf.gii'),
        iterfield=['in_file', 'hemi', 'suffix'],
        name='ds_surfs',
        run_without_submitting=True,
    )
    name_regs = pe.MapNode(
        Path2BIDS(), iterfield='in_file', name='name_regs', run_without_submitting=True
    )
    ds_regs = pe.MapNode(
        DerivativesDataSink(
            base_directory=output_dir,
            space='fsaverage',
            desc='reg',
            suffix='sphere',
            extension='.surf.gii',
        ),
        iterfield=['in_file', 'hemi'],
        name='ds_regs',
        run_without_submitting=True,
    )
    name_reg_fsLR = pe.MapNode(
        Path2BIDS(), iterfield='in_file', name='name_reg_fsLR', run_without_submitting=True
    )
    ds_reg_fsLR = pe.MapNode(
        DerivativesDataSink(
            base_directory=output_dir,
            space='dHCP' if surface_recon == 'mcribs' else 'fsLR',
            desc='reg',
            suffix='sphere',
            extension='.surf.gii',
        ),
        iterfield=['in_file', 'hemi'],
        name='ds_reg_fsLR',
        run_without_submitting=True,
    )
    # Morphometrics
    name_morphs = pe.MapNode(
        Path2BIDS(),
        iterfield='in_file',
        name='name_morphs',
        run_without_submitting=True,
    )
    ds_morphs = pe.MapNode(
        DerivativesDataSink(base_directory=output_dir, extension='.shape.gii'),
        iterfield=['in_file', 'hemi', 'suffix'],
        name='ds_morphs',
        run_without_submitting=True,
    )
    # Parcellations
    ds_anat_fsaseg = pe.Node(
        DerivativesDataSink(base_directory=output_dir, desc='aseg', suffix='dseg', compress=True),
        name='ds_anat_fsaseg',
        run_without_submitting=True,
    )
    ds_anat_fsparc = pe.Node(
        DerivativesDataSink(
            base_directory=output_dir, desc='aparcaseg', suffix='dseg', compress=True
        ),
        name='ds_anat_fsparc',
        run_without_submitting=True,
    )

    # fmt: off
    workflow.connect([
        (inputnode, lta2itk_fwd, [('anat2fsnative_xfm', 'in_xfms')]),
        (inputnode, lta2itk_inv, [('fsnative2anat_xfm', 'in_xfms')]),
        (inputnode, ds_anat_fsnative, [(source_files, 'source_file')]),
        (lta2itk_fwd, ds_anat_fsnative, [('out_xfm', 'in_file')]),
        (inputnode, ds_fsnative_anat, [(source_files, 'source_file')]),
        (lta2itk_inv, ds_fsnative_anat, [('out_xfm', 'in_file')]),
        (inputnode, name_surfs, [('surfaces', 'in_file')]),
        (inputnode, ds_surfs, [('surfaces', 'in_file'),
                               (source_files, 'source_file')]),
        (name_surfs, ds_surfs, [('hemi', 'hemi'),
                                ('suffix', 'suffix')]),
        (inputnode, name_regs, [('sphere_reg', 'in_file')]),
        (inputnode, ds_regs, [('sphere_reg', 'in_file'),
                              (source_files, 'source_file')]),
        (name_regs, ds_regs, [('hemi', 'hemi')]),
        (inputnode, name_reg_fsLR, [('sphere_reg_fsLR', 'in_file')]),
        (inputnode, ds_reg_fsLR, [('sphere_reg_fsLR', 'in_file'),
                                  (source_files, 'source_file')]),
        (name_reg_fsLR, ds_reg_fsLR, [('hemi', 'hemi')]),
        (inputnode, name_morphs, [('morphometrics', 'in_file')]),
        (inputnode, ds_morphs, [('morphometrics', 'in_file'),
                                (source_files, 'source_file')]),
        (name_morphs, ds_morphs, [('hemi', 'hemi'),
                                  ('suffix', 'suffix')]),
        (inputnode, ds_anat_fsaseg, [('anat_fs_aseg', 'in_file'),
                                     (source_files, 'source_file')]),
        (inputnode, ds_anat_fsparc, [('anat_fs_aparc', 'in_file'),
                                     (source_files, 'source_file')]),
    ])
    # fmt: on
    if cifti_output:
        ds_cifti_morph = pe.MapNode(
            DerivativesDataSink(
                base_directory=output_dir,
                suffix=['curv', 'sulc', 'thickness'],
                compress=False,
                space='fsLR',
            ),
            name='ds_cifti_morph',
            run_without_submitting=True,
            iterfield=['in_file', 'meta_dict', 'suffix'],
        )
        # fmt:off
        workflow.connect([
            (inputnode, ds_cifti_morph, [('cifti_morph', 'in_file'),
                                         (source_files, 'source_file'),
                                         ('cifti_density', 'density'),
                                         (('cifti_metadata', _read_jsons), 'meta_dict')])
        ])
        # fmt:on
    return workflow


def init_ds_seg_wf(
    *,
    output_dir: str,
    seg_type: str,
    extra_entities: dict | None = None,
):
    """
    Set up a battery of datasinks to store derivatives in the right location.

    Parameters
    ----------
    bids_root : :obj:`str`
        Root path of BIDS dataset
    output_dir : :obj:`str`
        Directory in which to save derivatives
    seg_type : :obj:`str`
        Type of segmentation (aseg, aparcaseg, etc)
    extra_entities : :obj:`dict` or None
        Additional entities to add to filename
    name : :obj:`str`
        Workflow name (default: ds_anat_segs_wf)

    Inputs
    ------
    in_seg
        Input segmentation, in native anatomical space
    source_files
        List of input anatomical images
    """
    workflow = Workflow(name=f'ds_{seg_type}_wf')

    inputnode = pe.Node(
        niu.IdentityInterface(fields=['source_files', 'in_seg']),
        name='inputnode',
    )

    extra_entities = extra_entities or {}

    ds_seg = pe.Node(
        DerivativesDataSink(
            base_directory=output_dir,
            desc=seg_type,
            suffix='dseg',
            compress=True,
            **extra_entities,
        ),
        name='ds_anat_fsaseg',
        run_without_submitting=True,
    )

    workflow.connect([
        (inputnode, ds_seg, [
            ('in_seg', 'in_file'),
            ('source_files', 'source_file'),
        ]),
    ])  # fmt:skip

    return workflow


def _set_tpl_res(space, resolution):
    if space in ('UNCInfant', 'Fischer344'):
        from nipype.interfaces.base import Undefined

        return Undefined
    try:
        return int(resolution)
    except ValueError:
        return 1


def _read_jsons(in_file):
    from json import loads
    from pathlib import Path

    return [loads(Path(f).read_text()) for f in in_file]



import argparse
from collections import defaultdict
import json
from pathlib import Path
import warnings
warnings.simplefilter("ignore", UserWarning)

import nibabel as nb
import nitransforms as nt
import numpy as np


DEFAULT_OUTPUT_SPACES = {"MNIInfant",}

CHECKMARK = u' \u2713'


def get_parser():

    # class AddSpaces(argparse.Action):
    #     _spaces = {"MNIInfant", }

    #     def __call__(self, parser, args, values, option_string=None):
    #         for value in values:
    #             if "cohort" in value:  # skip cohorts for now
    #                 continue
    #             self._spaces.add(value)
    #         setattr(args, self.dest, self._spaces)

    parser = argparse.ArgumentParser(description="Script to verify NiBabies outputs")
    parser.add_argument("output_dir", type=Path, help="Directory containing nibabies outputs")
    return parser


def main(argv=None):
    parser = get_parser()
    pargs = parser.parse_args(argv)
    test_output_layout(pargs.output_dir)
    for anats in pargs.output_dir.glob("sub-*/**/anat"):
        if not anats.is_dir():
            continue
        test_anatomicals(anats)

    for funcs in pargs.output_dir.glob("sub-*/**/func"):
        if not funcs.is_dir():
            continue
        test_functionals(funcs)
    print(f"No errors found for output directory {str(pargs.output_dir)}")


def test_output_layout(output_dir):
    assert output_dir.exists()
    dd = output_dir / 'dataset_description.json'
    assert dd.exists()
    assert 'NiBabies' in json.loads(dd.read_text())['Name']
    for ext in ('bib', 'html', 'md', 'tex'):
        assert (output_dir / "logs" / f"CITATION.{ext}").exists()


def test_anatomicals(anat_dir):
    print(f"** Checking anatomical derivatives ({str(anat_dir)}):")
    outputs = defaultdict(list)
    for fl in anat_dir.glob("*"):
        outtype = None
        if 'desc-brain_mask' in fl.name:
            outtype = 'masks'
        elif 'desc-preproc_T1w' in fl.name:
            outtype = 'preprocs'
        elif '_dseg.' in fl.name or '_probseg.' in fl.name:
            outtype = 'segs'
        elif '_xfm.' in fl.name:
            outtype = 'xfms'
        elif 'surf.gii' in fl.name:
            outtype = 'surfs'

        if outtype:
            outputs[outtype].append(fl)

    _check_masks(outputs['masks'])
    _check_segs(outputs['segs'])
    _check_xfms(outputs['xfms'])
    _check_surfs(outputs['surfs'])
    _check_t1w_preprocs(outputs['preprocs'])


def test_functionals(func_dir):
    print(f"** Checking functional derivatives ({str(func_dir)}):")
    outputs = defaultdict(list)
    for fl in func_dir.glob("*"):
        outtype = None
        if 'desc-brain_mask' in fl.name:
            outtype = 'masks'
        elif 'desc-preproc_bold' in fl.name:
            outtype = 'preprocs'
        elif '_boldref.' in fl.name:
            outtype = 'boldrefs'
        elif 'desc-aseg_dseg' in fl.name or '_probseg.' in fl.name:
            outtype = 'segs'
        elif '_xfm.' in fl.name:
            outtype = 'xfms'
        elif '.surf.gii' in fl.name:
            outtype = 'surfs'
        elif 'bold.dtseries' in fl.name:
            outtype = 'ciftis'

        if outtype:
            outputs[outtype].append(fl)

    _check_bold_preprocs(outputs['preprocs'])
    _check_boldrefs(outputs['boldrefs'])
    _check_masks(outputs['masks'])
    _check_segs(outputs['segs'])
    _check_xfms(outputs['xfms'])
    _check_surfs(outputs['surfs'])
    _check_ciftis(outputs['ciftis'])


def _check_masks(masks):
    for mask in masks:
        print(str(mask), end='')
        if mask.name.endswith('.json'):
            metadata = json.loads(mask.read_text())
            assert metadata
            # assert metadata["Type"] == "Brain"
        else:
            img = nb.load(mask)
            assert img.dataobj.dtype == np.uint8
            assert np.all(np.unique(img.dataobj) == [0, 1])
        print(CHECKMARK)


def _check_segs(segs):
    for seg in segs:
        print(str(seg), end='')
        img = nb.load(seg)
        if 'desc-aseg' in seg.name:
            assert img.dataobj.dtype == np.int16
            labels = set(np.unique(img.dataobj))
            # check for common subcortical labels
            subcor = {26, 58, 18, 54, 16, 11, 50, 8, 47, 28, 60, 17, 53, 13, 52, 12, 51, 10, 49}
            missing = subcor.difference(labels)
            if missing:
                print(f"Missing labels {missing}")
        elif 'desc-aparcaseg' in seg.name:
            assert img.dataobj.dtype == np.int16
            labels = set(np.unique(img.dataobj))
            # check for common cortical labels
            cort = {1000, 1007, 1022, 2000, 2007, 2022}
            missing = cort.difference(labels)
            if missing:
                print(f"Missing labels {missing}")
        elif '_probseg.' in seg.name:
            assert img.dataobj.dtype == np.float32
            assert np.max(img.dataobj) == 1
            assert np.min(img.dataobj) == 0
        print(CHECKMARK)


def _check_xfms(xfms):
    for xfm in xfms:
        print(str(xfm), end='')
        if xfm.name.endswith(".txt"):
            assert nt.linear.load(xfm, fmt='itk')
        elif xfm.name.endswith('.h5'):
            assert nt.manip.load(xfm)
        elif xfm.name.endswith('.json'):
            meta = json.loads(xfm.read_text())
            assert 'Sources' in meta
        else:
            raise NotImplementedError
        print(CHECKMARK)


def _check_surfs(surfs):
    for surf in surfs:
        print(str(surf), end='')
        if surf.name.endswith('.surf.gii'):
            img = nb.load(surf)
            assert img.numDA == 2
            da0, da1 = img.darrays
            assert da0.intent == 1008  # NIFTI_INTENT_POINTSET
            assert da1.intent == 1009  # NIFTI_INTENT_TRIANGLE
        print(CHECKMARK)


def _check_t1w_preprocs(preprocs):
    for preproc in preprocs:
        print(str(preproc), end='')
        if '.json' in preproc.name:
            metadata = json.loads(preproc.read_text())
            assert 'SkullStripped' in metadata
        else:
            img = nb.load(preproc)
            assert len(img.shape) == 3
        print(CHECKMARK)


def _check_bold_preprocs(preprocs):
    for preproc in preprocs:
        print(str(preproc), end='')
        if '.json' in preproc.name:
            metadata = json.loads(preproc.read_text())
            assert metadata['SkullStripped'] is False
        else:
            img = nb.load(preproc)
            assert len(img.shape) == 4
        print(CHECKMARK)


def _check_boldrefs(boldrefs):
    for boldref in boldrefs:
        print(str(boldref), end='')
        if '.json' in boldref.name:
            metadata = json.loads(boldref.read_text())
            assert 'Sources' in metadata
        else:
            bimg = nb.load(boldref)
            if len(bimg.shape) > 3:
                assert len(bimg.shape) == 4 and bimg.shape[3] == 1
        print(CHECKMARK)


def _check_ciftis(ciftis):
    for cifti in ciftis:
        print(str(cifti), end='')
        if '.json' in cifti.name:
            metadata = json.loads(cifti.read_text())
            assert metadata
        else:
            img = nb.load(cifti)
            assert len(img.shape) == 2
            matrix = img.header.matrix
            assert matrix.mapped_indices == [0, 1]
            series_map = matrix.get_index_map(0)
            bm_map = matrix.get_index_map(1)
            assert series_map.indices_map_to_data_type == 'CIFTI_INDEX_TYPE_SERIES'
            assert bm_map.indices_map_to_data_type == 'CIFTI_INDEX_TYPE_BRAIN_MODELS'
        print(CHECKMARK)


if __name__ == "__main__":
    main()



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# Copyright The NiPreps Developers <nipreps@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# We support and encourage derived works from this project, please read
# about our expectations at
#
#     https://www.nipreps.org/community/licensing/
#
"""Writing out derivative files."""

from __future__ import annotations

import numpy as np
from nipype.interfaces import utility as niu
from nipype.pipeline import engine as pe
from niworkflows.interfaces.fixes import FixHeaderApplyTransforms as ApplyTransforms
from niworkflows.utils.images import dseg_label

from nibabies import config
from nibabies._types import Anatomical
from nibabies.config import DEFAULT_DISMISS_ENTITIES, DEFAULT_MEMORY_MIN_GB, dismiss_echo
from nibabies.interfaces import DerivativesDataSink
from nibabies.interfaces.bids import BIDSURI


def prepare_timing_parameters(metadata: dict):
    """Convert initial timing metadata to post-realignment timing metadata

    In particular, SliceTiming metadata is invalid once STC or any realignment is applied,
    as a matrix of voxels no longer corresponds to an acquisition slice.
    Therefore, if SliceTiming is present in the metadata dictionary, and a sparse
    acquisition paradigm is detected, DelayTime or AcquisitionDuration must be derived to
    preserve the timing interpretation.

    Examples
    --------

    .. testsetup::

        >>> from unittest import mock

    If SliceTiming metadata is absent, then the only change is to note that
    STC has not been applied:

    >>> prepare_timing_parameters(dict(RepetitionTime=2))
    {'RepetitionTime': 2, 'SliceTimingCorrected': False}
    >>> prepare_timing_parameters(dict(RepetitionTime=2, DelayTime=0.5))
    {'RepetitionTime': 2, 'DelayTime': 0.5, 'SliceTimingCorrected': False}
    >>> prepare_timing_parameters(dict(VolumeTiming=[0.0, 1.0, 2.0, 5.0, 6.0, 7.0],
    ...                                AcquisitionDuration=1.0))  #doctest: +NORMALIZE_WHITESPACE
    {'VolumeTiming': [0.0, 1.0, 2.0, 5.0, 6.0, 7.0], 'AcquisitionDuration': 1.0,
     'SliceTimingCorrected': False}

    When SliceTiming is available and used, then ``SliceTimingCorrected`` is ``True``
    and the ``StartTime`` indicates a series offset.

    >>> with mock.patch("nibabies.config.workflow.ignore", []):
    ...     prepare_timing_parameters(dict(RepetitionTime=2, SliceTiming=[0.0, 0.2, 0.4, 0.6]))
    {'RepetitionTime': 2, 'SliceTimingCorrected': True, 'DelayTime': 1.2, 'StartTime': 0.3}
    >>> with mock.patch("nibabies.config.workflow.ignore", []):
    ...     prepare_timing_parameters(
    ...         dict(VolumeTiming=[0.0, 1.0, 2.0, 5.0, 6.0, 7.0],
    ...              SliceTiming=[0.0, 0.2, 0.4, 0.6, 0.8]))  #doctest: +NORMALIZE_WHITESPACE
    {'VolumeTiming': [0.0, 1.0, 2.0, 5.0, 6.0, 7.0], 'SliceTimingCorrected': True,
     'AcquisitionDuration': 1.0, 'StartTime': 0.4}

    When SliceTiming is available and not used, then ``SliceTimingCorrected`` is ``False``
    and TA is indicated with ``DelayTime`` or ``AcquisitionDuration``.

    >>> with mock.patch("nibabies.config.workflow.ignore", ["slicetiming"]):
    ...     prepare_timing_parameters(dict(RepetitionTime=2, SliceTiming=[0.0, 0.2, 0.4, 0.6]))
    {'RepetitionTime': 2, 'SliceTimingCorrected': False, 'DelayTime': 1.2}
    >>> with mock.patch("nibabies.config.workflow.ignore", ["slicetiming"]):
    ...     prepare_timing_parameters(
    ...         dict(VolumeTiming=[0.0, 1.0, 2.0, 5.0, 6.0, 7.0],
    ...              SliceTiming=[0.0, 0.2, 0.4, 0.6, 0.8]))  #doctest: +NORMALIZE_WHITESPACE
    {'VolumeTiming': [0.0, 1.0, 2.0, 5.0, 6.0, 7.0], 'SliceTimingCorrected': False,
     'AcquisitionDuration': 1.0}

    If SliceTiming metadata is present but empty, then treat it as missing:

    >>> with mock.patch("nibabies.config.workflow.ignore", []):
    ...     prepare_timing_parameters(dict(RepetitionTime=2, SliceTiming=[]))
    {'RepetitionTime': 2, 'SliceTimingCorrected': False}
    >>> with mock.patch("nibabies.config.workflow.ignore", []):
    ...     prepare_timing_parameters(dict(RepetitionTime=2, SliceTiming=[0.0]))
    {'RepetitionTime': 2, 'SliceTimingCorrected': False}
    """
    timing_parameters = {
        key: metadata[key]
        for key in (
            'RepetitionTime',
            'VolumeTiming',
            'DelayTime',
            'AcquisitionDuration',
            'SliceTiming',
        )
        if key in metadata
    }

    # Treat SliceTiming of [] or length 1 as equivalent to missing and remove it in any case
    slice_timing = timing_parameters.pop('SliceTiming', [])

    run_stc = len(slice_timing) > 1 and 'slicetiming' not in config.workflow.ignore
    timing_parameters['SliceTimingCorrected'] = run_stc

    if len(slice_timing) > 1:
        st = sorted(slice_timing)
        TA = st[-1] + (st[1] - st[0])  # Final slice onset + slice duration
        # For constant TR paradigms, use DelayTime
        if 'RepetitionTime' in timing_parameters:
            TR = timing_parameters['RepetitionTime']
            if not np.isclose(TR, TA) and TA < TR:
                timing_parameters['DelayTime'] = TR - TA
        # For variable TR paradigms, use AcquisitionDuration
        elif 'VolumeTiming' in timing_parameters:
            timing_parameters['AcquisitionDuration'] = TA

        if run_stc:
            first, last = st[0], st[-1]
            frac = config.workflow.slice_time_ref
            tzero = float(np.round(first + frac * (last - first), 3))
            timing_parameters['StartTime'] = tzero

    return timing_parameters


def init_func_fit_reports_wf(
    *,
    reference_anat: Anatomical,
    sdc_correction: bool,
    output_dir: str,
    name='func_fit_reports_wf',
) -> pe.Workflow:
    """
    Set up a battery of datasinks to store reports in the right location.

    Parameters
    ----------
    reference_anat : :obj:`str`
        Reference anatomical image type (T1w, T2w)
    sdc_correction : :obj:`bool`
        SDC was performed
    output_dir : :obj:`str`
        Directory in which to save derivatives
    name : :obj:`str`
        Workflow name (default: func_fit_reports_wf)

    Inputs
    ------
    source_file
        Input BOLD images
    subject_dir
        FreeSurfer SUBJECTS_DIR
    subject_id
        FreeSurfer subject ID
    anat_preproc
        The anatomical reference map, which is calculated as the average of bias-corrected
        and preprocessed anatomical images, defining the anatomical space.
    anat_dseg
        Segmentation in anatomical reference space
    anat_mask
        Brain (binary) mask estimated by brain extraction.
    """
    from nireports.interfaces.reporting.base import (
        SimpleBeforeAfterRPT as SimpleBeforeAfter,
    )
    from sdcflows.interfaces.reportlets import FieldmapReportlet

    workflow = pe.Workflow(name=name)

    inputfields = [
        'source_file',
        'sdc_boldref',
        'coreg_boldref',
        'bold_mask',
        'boldref2anat_xfm',
        'boldref2fmap_xfm',
        'anat_preproc',
        'anat_mask',
        'anat_dseg',
        'fieldmap',
        'fmap_ref',
        # May be missing
        'subject_id',
        'subjects_dir',
        # Report snippets
        'summary_report',
        'validation_report',
    ]
    inputnode = pe.Node(niu.IdentityInterface(fields=inputfields), name='inputnode')

    ds_summary = pe.Node(
        DerivativesDataSink(
            base_directory=output_dir,
            desc='summary',
            datatype='figures',
            dismiss_entities=DEFAULT_DISMISS_ENTITIES,
        ),
        name='ds_report_summary',
        run_without_submitting=True,
        mem_gb=config.DEFAULT_MEMORY_MIN_GB,
    )

    ds_validation = pe.Node(
        DerivativesDataSink(
            base_directory=output_dir,
            desc='validation',
            datatype='figures',
            dismiss_entities=DEFAULT_DISMISS_ENTITIES,
        ),
        name='ds_report_validation',
        run_without_submitting=True,
        mem_gb=config.DEFAULT_MEMORY_MIN_GB,
    )

    # Resample anatomical references into BOLD space for plotting
    anat_boldref = pe.Node(
        ApplyTransforms(
            dimension=3,
            default_value=0,
            float=True,
            invert_transform_flags=[True],
            interpolation='LanczosWindowedSinc',
        ),
        name='anat_boldref',
        mem_gb=1,
    )

    anat_wm = pe.Node(
        niu.Function(function=dseg_label),
        name='anat_wm',
        mem_gb=DEFAULT_MEMORY_MIN_GB,
    )
    # TODO: Verify this is consistent with T1/T2
    anat_wm.inputs.label = 2  # BIDS default is WM=2

    boldref_wm = pe.Node(
        ApplyTransforms(
            dimension=3,
            default_value=0,
            invert_transform_flags=[True],
            interpolation='NearestNeighbor',
        ),
        name='boldref_wm',
        mem_gb=1,
    )

    workflow.connect([
        (inputnode, ds_summary, [
            ('source_file', 'source_file'),
            ('summary_report', 'in_file'),
        ]),
        (inputnode, ds_validation, [
            ('source_file', 'source_file'),
            ('validation_report', 'in_file'),
        ]),
        (inputnode, anat_boldref, [
            ('anat_preproc', 'input_image'),
            ('coreg_boldref', 'reference_image'),
            ('boldref2anat_xfm', 'transforms'),
        ]),
        (inputnode, anat_wm, [('anat_dseg', 'in_seg')]),
        (inputnode, boldref_wm, [
            ('coreg_boldref', 'reference_image'),
            ('boldref2anat_xfm', 'transforms'),
        ]),
        (anat_wm, boldref_wm, [('out', 'input_image')]),
    ])  # fmt:skip

    # Reportlets follow the structure of init_bold_fit_wf stages
    # - SDC1:
    #       Before: Pre-SDC boldref
    #       After: Fieldmap reference resampled on boldref
    #       Three-way: Fieldmap resampled on boldref
    # - SDC2:
    #       Before: Pre-SDC boldref with white matter mask
    #       After: Post-SDC boldref with white matter mask
    # - EPI-anat registration:
    #       Before: anat brain with white matter mask
    #       After: Resampled boldref with white matter mask

    if sdc_correction:
        fmapref_boldref = pe.Node(
            ApplyTransforms(
                dimension=3,
                default_value=0,
                float=True,
                invert_transform_flags=[True],
                interpolation='LanczosWindowedSinc',
            ),
            name='fmapref_boldref',
            mem_gb=1,
        )

        # SDC1
        sdcreg_report = pe.Node(
            FieldmapReportlet(
                reference_label='BOLD reference',
                moving_label='Fieldmap reference',
                show='both',
            ),
            name='sdecreg_report',
            mem_gb=0.1,
        )

        ds_sdcreg_report = pe.Node(
            DerivativesDataSink(
                base_directory=output_dir,
                desc='fmapCoreg',
                suffix='bold',
                datatype='figures',
                dismiss_entities=DEFAULT_DISMISS_ENTITIES,
            ),
            name='ds_sdcreg_report',
        )

        # SDC2
        sdc_report = pe.Node(
            SimpleBeforeAfter(
                before_label='Distorted',
                after_label='Corrected',
                dismiss_affine=True,
            ),
            name='sdc_report',
            mem_gb=0.1,
        )

        ds_sdc_report = pe.Node(
            DerivativesDataSink(
                base_directory=output_dir,
                desc='sdc',
                suffix='bold',
                datatype='figures',
                dismiss_entities=DEFAULT_DISMISS_ENTITIES,
            ),
            name='ds_sdc_report',
        )

        workflow.connect([
            (inputnode, fmapref_boldref, [
                ('fmap_ref', 'input_image'),
                ('coreg_boldref', 'reference_image'),
                ('boldref2fmap_xfm', 'transforms'),
            ]),
            (inputnode, sdcreg_report, [
                ('sdc_boldref', 'reference'),
                ('fieldmap', 'fieldmap'),
                ('bold_mask', 'mask'),
            ]),
            (fmapref_boldref, sdcreg_report, [('output_image', 'moving')]),
            (inputnode, ds_sdcreg_report, [('source_file', 'source_file')]),
            (sdcreg_report, ds_sdcreg_report, [('out_report', 'in_file')]),
            (inputnode, sdc_report, [
                ('sdc_boldref', 'before'),
                ('coreg_boldref', 'after'),
            ]),
            (boldref_wm, sdc_report, [('output_image', 'wm_seg')]),
            (inputnode, ds_sdc_report, [('source_file', 'source_file')]),
            (sdc_report, ds_sdc_report, [('out_report', 'in_file')]),
        ])  # fmt:skip

    # EPI-anat registration
    # Resample anat image onto EPI-space

    epi_anat_report = pe.Node(
        SimpleBeforeAfter(
            before_label=reference_anat,
            after_label='EPI',
            dismiss_affine=True,
        ),
        name='epi_anat_report',
        mem_gb=0.1,
    )

    ds_epi_anat_report = pe.Node(
        DerivativesDataSink(
            base_directory=output_dir,
            desc='coreg',
            suffix='bold',
            datatype='figures',
            dismiss_entities=DEFAULT_DISMISS_ENTITIES,
        ),
        name='ds_epi_anat_report',
    )

    workflow.connect([
        (inputnode, epi_anat_report, [('coreg_boldref', 'after')]),
        (anat_boldref, epi_anat_report, [('output_image', 'before')]),
        (boldref_wm, epi_anat_report, [('output_image', 'wm_seg')]),
        (inputnode, ds_epi_anat_report, [('source_file', 'source_file')]),
        (epi_anat_report, ds_epi_anat_report, [('out_report', 'in_file')]),
    ])  # fmt:skip

    return workflow


def init_ds_boldref_wf(
    *,
    output_dir,
    desc: str,
    name='ds_boldref_wf',
) -> pe.Workflow:
    workflow = pe.Workflow(name=name)

    inputnode = pe.Node(
        niu.IdentityInterface(fields=['source_files', 'boldref']),
        name='inputnode',
    )
    outputnode = pe.Node(niu.IdentityInterface(fields=['boldref']), name='outputnode')

    sources = pe.Node(
        BIDSURI(
            numinputs=1,
            dataset_links=config.execution.dataset_links,
            out_dir=str(config.execution.output_dir.absolute()),
        ),
        name='sources',
    )

    ds_boldref = pe.Node(
        DerivativesDataSink(
            base_directory=output_dir,
            desc=desc,
            suffix='boldref',
            compress=True,
            dismiss_entities=DEFAULT_DISMISS_ENTITIES,
        ),
        name='ds_boldref',
        run_without_submitting=True,
    )

    # fmt:off
    workflow.connect([
        (inputnode, sources, [('source_files', 'in1')]),
        (inputnode, ds_boldref, [('boldref', 'in_file'),
                                 ('source_files', 'source_file')]),
        (sources, ds_boldref, [('out', 'Sources')]),
        (ds_boldref, outputnode, [('out_file', 'boldref')]),
    ])
    # fmt:on

    return workflow


def init_ds_registration_wf(
    *,
    output_dir: str,
    source: str,
    dest: str,
    name: str,
) -> pe.Workflow:
    workflow = pe.Workflow(name=name)

    inputnode = pe.Node(
        niu.IdentityInterface(fields=['source_files', 'xform']),
        name='inputnode',
    )
    outputnode = pe.Node(niu.IdentityInterface(fields=['xform']), name='outputnode')

    sources = pe.Node(
        BIDSURI(
            numinputs=1,
            dataset_links=config.execution.dataset_links,
            out_dir=str(config.execution.output_dir.absolute()),
        ),
        name='sources',
    )

    ds_xform = pe.Node(
        DerivativesDataSink(
            base_directory=output_dir,
            mode='image',
            suffix='xfm',
            extension='.txt',
            dismiss_entities=dismiss_echo(['part']),
            **{'from': source, 'to': dest},
        ),
        name='ds_xform',
        run_without_submitting=True,
        mem_gb=DEFAULT_MEMORY_MIN_GB,
    )

    # fmt:off
    workflow.connect([
        (inputnode, sources, [('source_files', 'in1')]),
        (inputnode, ds_xform, [('xform', 'in_file'),
                               ('source_files', 'source_file')]),
        (sources, ds_xform, [('out', 'Sources')]),
        (ds_xform, outputnode, [('out_file', 'xform')]),
    ])
    # fmt:on

    return workflow


def init_ds_hmc_wf(
    *,
    output_dir,
    name='ds_hmc_wf',
) -> pe.Workflow:
    workflow = pe.Workflow(name=name)

    inputnode = pe.Node(
        niu.IdentityInterface(fields=['source_files', 'xforms']),
        name='inputnode',
    )
    outputnode = pe.Node(niu.IdentityInterface(fields=['xforms']), name='outputnode')

    sources = pe.Node(
        BIDSURI(
            numinputs=1,
            dataset_links=config.execution.dataset_links,
            out_dir=str(config.execution.output_dir.absolute()),
        ),
        name='sources',
    )

    ds_xforms = pe.Node(
        DerivativesDataSink(
            base_directory=output_dir,
            desc='hmc',
            suffix='xfm',
            extension='.txt',
            compress=True,
            dismiss_entities=DEFAULT_DISMISS_ENTITIES,
            **{'from': 'orig', 'to': 'boldref'},
        ),
        name='ds_xforms',
        run_without_submitting=True,
    )

    # fmt:off
    workflow.connect([
        (inputnode, sources, [('source_files', 'in1')]),
        (inputnode, ds_xforms, [('xforms', 'in_file'),
                                ('source_files', 'source_file')]),
        (sources, ds_xforms, [('out', 'Sources')]),
        (ds_xforms, outputnode, [('out_file', 'xforms')]),
    ])
    # fmt:on

    return workflow


def init_ds_bold_native_wf(
    *,
    bids_root: str,
    output_dir: str,
    multiecho: bool,
    bold_output: bool,
    echo_output: bool,
    all_metadata: list[dict],
    name='ds_bold_native_wf',
) -> pe.Workflow:
    metadata = all_metadata[0]
    timing_parameters = prepare_timing_parameters(metadata)

    workflow = pe.Workflow(name=name)
    inputnode = pe.Node(
        niu.IdentityInterface(
            fields=[
                'source_files',
                'bold',
                'bold_mask',
                'bold_echos',
                't2star',
                # Transforms previously used to generate the outputs
                'motion_xfm',
                'boldref2fmap_xfm',
            ]
        ),
        name='inputnode',
    )

    sources = pe.Node(
        BIDSURI(
            numinputs=3,
            dataset_links=config.execution.dataset_links,
            out_dir=str(config.execution.output_dir.absolute()),
        ),
        name='sources',
    )
    workflow.connect([
        (inputnode, sources, [
            ('source_files', 'in1'),
            ('motion_xfm', 'in2'),
            ('boldref2fmap_xfm', 'in3'),
        ]),
    ])  # fmt:skip

    # Masks should be output if any other derivatives are output
    ds_bold_mask = pe.Node(
        DerivativesDataSink(
            base_directory=output_dir,
            desc='brain',
            suffix='mask',
            compress=True,
            dismiss_entities=DEFAULT_DISMISS_ENTITIES,
        ),
        name='ds_bold_mask',
        run_without_submitting=True,
        mem_gb=DEFAULT_MEMORY_MIN_GB,
    )
    workflow.connect([
        (inputnode, ds_bold_mask, [
            ('source_files', 'source_file'),
            ('bold_mask', 'in_file'),
        ]),
        (sources, ds_bold_mask, [('out', 'Sources')]),
    ])  # fmt:skip

    if bold_output:
        ds_bold = pe.Node(
            DerivativesDataSink(
                base_directory=output_dir,
                desc='preproc',
                compress=True,
                SkullStripped=multiecho,
                TaskName=metadata.get('TaskName'),
                dismiss_entities=DEFAULT_DISMISS_ENTITIES,
                **timing_parameters,
            ),
            name='ds_bold',
            mem_gb=DEFAULT_MEMORY_MIN_GB,
        )
        workflow.connect([
            (inputnode, ds_bold, [
                ('source_files', 'source_file'),
                ('bold', 'in_file'),
            ]),
            (sources, ds_bold, [('out', 'Sources')]),
        ])  # fmt:skip

    if bold_output and multiecho:
        t2star_meta = {
            'Units': 's',
            'EstimationReference': 'doi:10.1002/mrm.20900',
            'EstimationAlgorithm': 'monoexponential decay model',
        }
        ds_t2star = pe.Node(
            DerivativesDataSink(
                base_directory=output_dir,
                space='boldref',
                suffix='T2starmap',
                compress=True,
                dismiss_entities=DEFAULT_DISMISS_ENTITIES,
                **t2star_meta,
            ),
            name='ds_t2star_bold',
            run_without_submitting=True,
            mem_gb=DEFAULT_MEMORY_MIN_GB,
        )
        workflow.connect([
            (inputnode, ds_t2star, [
                ('source_files', 'source_file'),
                ('t2star', 'in_file'),
            ]),
            (sources, ds_t2star, [('out', 'Sources')]),
        ])  # fmt:skip

    if echo_output:
        ds_bold_echos = pe.MapNode(
            DerivativesDataSink(
                base_directory=output_dir,
                desc='preproc',
                compress=True,
                SkullStripped=False,
                TaskName=metadata.get('TaskName'),
                **timing_parameters,
            ),
            iterfield=['source_file', 'in_file', 'meta_dict'],
            name='ds_bold_echos',
            run_without_submitting=True,
            mem_gb=DEFAULT_MEMORY_MIN_GB,
        )
        ds_bold_echos.inputs.meta_dict = [{'EchoTime': md['EchoTime']} for md in all_metadata]
        workflow.connect([
            (inputnode, ds_bold_echos, [
                ('source_files', 'source_file'),
                ('bold_echos', 'in_file'),
            ]),
        ])  # fmt:skip

    return workflow


def init_ds_volumes_wf(
    *,
    bids_root: str,
    output_dir: str,
    multiecho: bool,
    metadata: list[dict],
    name='ds_volumes_wf',
) -> pe.Workflow:
    timing_parameters = prepare_timing_parameters(metadata)

    workflow = pe.Workflow(name=name)
    inputnode = pe.Node(
        niu.IdentityInterface(
            fields=[
                'source_files',
                'ref_file',
                'bold',  # Resampled into target space
                'bold_mask',  # boldref space
                'bold_ref',  # boldref space
                't2star',  # boldref space
                'template',  # target reference image from original transform
                # Anatomical
                'boldref2anat_xfm',
                # Template
                'anat2std_xfm',
                # Entities
                'space',
                'cohort',
                'resolution',
                # Transforms previously used to generate the outputs
                'motion_xfm',
                'boldref2fmap_xfm',
            ]
        ),
        name='inputnode',
    )

    sources = pe.Node(
        BIDSURI(
            numinputs=6,
            dataset_links=config.execution.dataset_links,
            out_dir=str(config.execution.output_dir.absolute()),
        ),
        name='sources',
    )
    boldref2target = pe.Node(niu.Merge(2), name='boldref2target')

    # BOLD is pre-resampled
    ds_bold = pe.Node(
        DerivativesDataSink(
            base_directory=output_dir,
            desc='preproc',
            compress=True,
            SkullStripped=multiecho,
            TaskName=metadata.get('TaskName'),
            dismiss_entities=DEFAULT_DISMISS_ENTITIES,
            **timing_parameters,
        ),
        name='ds_bold',
        mem_gb=DEFAULT_MEMORY_MIN_GB,
    )
    workflow.connect([
        (inputnode, sources, [
            ('source_files', 'in1'),
            ('motion_xfm', 'in2'),
            ('boldref2fmap_xfm', 'in3'),
            ('boldref2anat_xfm', 'in4'),
            ('anat2std_xfm', 'in5'),
            ('template', 'in6'),
        ]),
        (inputnode, boldref2target, [
            # Note that ANTs expects transforms in target-to-source order
            # Reverse this for nitransforms-based resamplers
            ('anat2std_xfm', 'in1'),
            ('boldref2anat_xfm', 'in2'),
        ]),
        (inputnode, ds_bold, [
            ('source_files', 'source_file'),
            ('bold', 'in_file'),
            ('space', 'space'),
            ('cohort', 'cohort'),
            ('resolution', 'resolution'),
        ]),
        (sources, ds_bold, [('out', 'Sources')]),
    ])  # fmt:skip

    resample_ref = pe.Node(
        ApplyTransforms(
            dimension=3,
            default_value=0,
            float=True,
            interpolation='LanczosWindowedSinc',
        ),
        name='resample_ref',
    )
    resample_mask = pe.Node(ApplyTransforms(interpolation='MultiLabel'), name='resample_mask')
    resamplers = [resample_ref, resample_mask]

    workflow.connect([
        (inputnode, resample_ref, [('bold_ref', 'input_image')]),
        (inputnode, resample_mask, [('bold_mask', 'input_image')]),
    ])  # fmt:skip

    ds_ref = pe.Node(
        DerivativesDataSink(
            base_directory=output_dir,
            suffix='boldref',
            compress=True,
            dismiss_entities=DEFAULT_DISMISS_ENTITIES,
        ),
        name='ds_ref',
        run_without_submitting=True,
        mem_gb=DEFAULT_MEMORY_MIN_GB,
    )
    ds_mask = pe.Node(
        DerivativesDataSink(
            base_directory=output_dir,
            desc='brain',
            suffix='mask',
            compress=True,
            dismiss_entities=DEFAULT_DISMISS_ENTITIES,
        ),
        name='ds_mask',
        run_without_submitting=True,
        mem_gb=DEFAULT_MEMORY_MIN_GB,
    )
    datasinks = [ds_ref, ds_mask]

    if multiecho:
        t2star_meta = {
            'Units': 's',
            'EstimationReference': 'doi:10.1002/mrm.20900',
            'EstimationAlgorithm': 'monoexponential decay model',
        }
        resample_t2star = pe.Node(
            ApplyTransforms(
                dimension=3,
                default_value=0,
                float=True,
                interpolation='LanczosWindowedSinc',
            ),
            name='resample_t2star',
        )
        ds_t2star = pe.Node(
            DerivativesDataSink(
                base_directory=output_dir,
                suffix='T2starmap',
                compress=True,
                dismiss_entities=DEFAULT_DISMISS_ENTITIES,
                **t2star_meta,
            ),
            name='ds_t2star_std',
            run_without_submitting=True,
            mem_gb=DEFAULT_MEMORY_MIN_GB,
        )
        resamplers.append(resample_t2star)
        datasinks.append(ds_t2star)

        workflow.connect([(inputnode, resample_t2star, [('t2star', 'input_image')])])

    workflow.connect(
        [
            (inputnode, resampler, [('ref_file', 'reference_image')])
            for resampler in resamplers
        ] + [
            (boldref2target, resampler, [('out', 'transforms')])
            for resampler in resamplers
        ] + [
            (inputnode, datasink, [
                ('source_files', 'source_file'),
                ('space', 'space'),
                ('cohort', 'cohort'),
                ('resolution', 'resolution'),
            ])
            for datasink in datasinks
        ] + [
            (sources, datasink, [('out', 'Sources')])
            for datasink in datasinks
        ] + [
            (resampler, datasink, [('output_image', 'in_file')])
            for resampler, datasink in zip(resamplers, datasinks, strict=False)
        ]
    )  # fmt:skip

    return workflow


def init_bold_preproc_report_wf(
    mem_gb: float,
    reportlets_dir: str,
    name: str = 'bold_preproc_report_wf',
):
    """
    Generate a visual report.

    This workflow generates and saves a reportlet showing the effect of resampling
    the BOLD signal using the standard deviation maps.

    Workflow Graph
        .. workflow::
            :graph2use: orig
            :simple_form: yes

            from nibabies.workflows.bold.resampling import init_bold_preproc_report_wf
            wf = init_bold_preproc_report_wf(mem_gb=1, reportlets_dir='.')

    Parameters
    ----------
    mem_gb : :obj:`float`
        Size of BOLD file in GB
    reportlets_dir : :obj:`str`
        Directory in which to save reportlets
    name : :obj:`str`, optional
        Workflow name (default: bold_preproc_report_wf)

    Inputs
    ------
    in_pre
        BOLD time-series, before resampling
    in_post
        BOLD time-series, after resampling
    name_source
        BOLD series NIfTI file
        Used to recover original information lost during processing

    """
    from nipype.algorithms.confounds import TSNR
    from nireports.interfaces.reporting.base import SimpleBeforeAfterRPT
    from niworkflows.engine.workflows import LiterateWorkflow as Workflow

    from ...interfaces import DerivativesDataSink

    workflow = Workflow(name=name)

    inputnode = pe.Node(
        niu.IdentityInterface(fields=['in_pre', 'in_post', 'name_source']), name='inputnode'
    )

    pre_tsnr = pe.Node(TSNR(), name='pre_tsnr', mem_gb=mem_gb * 4.5)
    pos_tsnr = pe.Node(TSNR(), name='pos_tsnr', mem_gb=mem_gb * 4.5)

    bold_rpt = pe.Node(SimpleBeforeAfterRPT(), name='bold_rpt', mem_gb=0.1)
    ds_report_bold = pe.Node(
        DerivativesDataSink(
            base_directory=reportlets_dir,
            desc='preproc',
            datatype='figures',
            dismiss_entities=DEFAULT_DISMISS_ENTITIES,
        ),
        name='ds_report_bold',
        mem_gb=DEFAULT_MEMORY_MIN_GB,
        run_without_submitting=True,
    )
    # fmt:off
    workflow.connect([
        (inputnode, ds_report_bold, [('name_source', 'source_file')]),
        (inputnode, pre_tsnr, [('in_pre', 'in_file')]),
        (inputnode, pos_tsnr, [('in_post', 'in_file')]),
        (pre_tsnr, bold_rpt, [('stddev_file', 'before')]),
        (pos_tsnr, bold_rpt, [('stddev_file', 'after')]),
        (bold_rpt, ds_report_bold, [('out_report', 'in_file')]),
    ])
    # fmt:on

    return workflow



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# Copyright The NiPreps Developers <nipreps@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# We support and encourage derived works from this project, please read
# about our expectations at
#
#     https://www.nipreps.org/community/licensing/
#
"""
Resampling workflows
++++++++++++++++++++

.. autofunction:: init_bold_surf_wf
.. autofunction:: init_bold_fsLR_resampling_wf
.. autofunction:: init_bold_grayords_wf
.. autofunction:: init_goodvoxels_bold_mask_wf

"""

from __future__ import annotations

import typing as ty

from nipype import Function
from nipype.interfaces import freesurfer as fs
from nipype.interfaces import fsl
from nipype.interfaces import utility as niu
from nipype.pipeline import engine as pe
from niworkflows.interfaces.fixes import FixHeaderApplyTransforms as ApplyTransforms
from niworkflows.interfaces.freesurfer import MedialNaNs
from niworkflows.interfaces.workbench import MetricDilate, MetricMask, MetricResample

from nibabies import config
from nibabies.config import DEFAULT_DISMISS_ENTITIES, DEFAULT_MEMORY_MIN_GB
from nibabies.interfaces.bids import BIDSURI
from nibabies.workflows.bold.outputs import prepare_timing_parameters


def init_bold_surf_wf(
    *,
    mem_gb: float,
    surface_spaces: list[str],
    medial_surface_nan: bool,
    metadata: dict,
    output_dir: str,
    name: str = 'bold_surf_wf',
):
    """
    Sample functional images to FreeSurfer surfaces.

    For each vertex, the cortical ribbon is sampled at six points (spaced 20% of thickness apart)
    and averaged.

    Outputs are in GIFTI format.

    Workflow Graph
        .. workflow::
            :graph2use: colored
            :simple_form: yes

            from nibabies.workflows.bold import init_bold_surf_wf
            wf = init_bold_surf_wf(mem_gb=0.1,
                                   surface_spaces=["fsnative", "fsaverage5"],
                                   medial_surface_nan=False,
                                   metadata={},
                                   output_dir='.',
                                   )

    Parameters
    ----------
    surface_spaces : :obj:`list`
        List of FreeSurfer surface-spaces (either ``fsaverage{3,4,5,6,}`` or ``fsnative``)
        the functional images are to be resampled to.
        For ``fsnative``, images will be resampled to the individual subject's
        native surface.
    medial_surface_nan : :obj:`bool`
        Replace medial wall values with NaNs on functional GIFTI files

    Inputs
    ------
    source_file
        Original BOLD series
    sources
        List of files used to create the output files.
    bold_anat
        Motion-corrected BOLD series in T1 space
    subjects_dir
        FreeSurfer SUBJECTS_DIR
    subject_id
        FreeSurfer subject ID
    fsnative2anat_xfm
        ITK-style affine matrix translating from FreeSurfer-conformed subject space to anatomical

    Outputs
    -------
    surfaces
        BOLD series, resampled to FreeSurfer surfaces

    """
    from nipype.interfaces.io import FreeSurferSource
    from niworkflows.engine.workflows import LiterateWorkflow as Workflow
    from niworkflows.interfaces.nitransforms import ConcatenateXFMs
    from niworkflows.interfaces.surf import GiftiSetAnatomicalStructure

    from nibabies.interfaces import DerivativesDataSink

    timing_parameters = prepare_timing_parameters(metadata)

    workflow = Workflow(name=name)
    workflow.__desc__ = """\
The BOLD time-series were resampled onto the following surfaces
(FreeSurfer reconstruction nomenclature):
{out_spaces}.
""".format(out_spaces=', '.join([f'*{s}*' for s in surface_spaces]))

    inputnode = pe.Node(
        niu.IdentityInterface(
            fields=[
                'source_file',
                'sources',
                'bold_anat',
                'subject_id',
                'subjects_dir',
                'fsnative2anat_xfm',
            ]
        ),
        name='inputnode',
    )
    itersource = pe.Node(niu.IdentityInterface(fields=['target']), name='itersource')
    itersource.iterables = [('target', surface_spaces)]

    surfs_sources = pe.Node(
        BIDSURI(
            numinputs=1,
            dataset_links=config.execution.dataset_links,
            out_dir=output_dir,
        ),
        name='surfs_sources',
    )

    get_fsnative = pe.Node(FreeSurferSource(), name='get_fsnative', run_without_submitting=True)

    def select_target(subject_id, space):
        """Get the target subject ID, given a source subject ID and a target space."""
        return subject_id if space == 'fsnative' else space

    targets = pe.Node(
        niu.Function(function=select_target),
        name='targets',
        run_without_submitting=True,
        mem_gb=DEFAULT_MEMORY_MIN_GB,
    )

    itk2lta = pe.Node(
        ConcatenateXFMs(out_fmt='fs', inverse=True), name='itk2lta', run_without_submitting=True
    )
    sampler = pe.MapNode(
        fs.SampleToSurface(
            interp_method='trilinear',
            out_type='gii',
            override_reg_subj=True,
            sampling_method='average',
            sampling_range=(0, 1, 0.2),
            sampling_units='frac',
        ),
        iterfield=['hemi'],
        name='sampler',
        mem_gb=mem_gb * 3,
    )
    sampler.inputs.hemi = ['lh', 'rh']

    update_metadata = pe.MapNode(
        GiftiSetAnatomicalStructure(),
        iterfield=['in_file'],
        name='update_metadata',
        mem_gb=DEFAULT_MEMORY_MIN_GB,
    )

    ds_bold_surfs = pe.MapNode(
        DerivativesDataSink(
            base_directory=output_dir,
            extension='.func.gii',
            dismiss_entities=DEFAULT_DISMISS_ENTITIES,
            TaskName=metadata.get('TaskName'),
            **timing_parameters,
        ),
        iterfield=['in_file', 'hemi'],
        name='ds_bold_surfs',
        run_without_submitting=True,
        mem_gb=DEFAULT_MEMORY_MIN_GB,
    )
    ds_bold_surfs.inputs.hemi = ['L', 'R']

    workflow.connect([
        (inputnode, get_fsnative, [
            ('subject_id', 'subject_id'),
            ('subjects_dir', 'subjects_dir')
        ]),
        (inputnode, targets, [('subject_id', 'subject_id')]),
        (inputnode, itk2lta, [
            ('bold_anat', 'moving'),
            ('fsnative2anat_xfm', 'in_xfms'),
        ]),
        # Use brain > T1 to ensure all recon methods have it
        (get_fsnative, itk2lta, [('brain', 'reference')]),
        (inputnode, sampler, [
            ('subjects_dir', 'subjects_dir'),
            ('subject_id', 'subject_id'),
            ('bold_anat', 'source_file'),
        ]),
        (itersource, targets, [('target', 'space')]),
        (itk2lta, sampler, [('out_inv', 'reg_file')]),
        (targets, sampler, [('out', 'target_subject')]),
        (inputnode, ds_bold_surfs, [('source_file', 'source_file')]),
        (inputnode, surfs_sources, [('sources', 'in1')]),
        (surfs_sources, ds_bold_surfs, [('out', 'Sources')]),
        (itersource, ds_bold_surfs, [('target', 'space')]),
        (update_metadata, ds_bold_surfs, [('out_file', 'in_file')]),
    ])  # fmt:skip

    # Refine if medial vertices should be NaNs
    medial_nans = pe.MapNode(
        MedialNaNs(), iterfield=['in_file'], name='medial_nans', mem_gb=DEFAULT_MEMORY_MIN_GB
    )

    if medial_surface_nan:
        workflow.connect([
            (inputnode, medial_nans, [('subjects_dir', 'subjects_dir')]),
            (sampler, medial_nans, [('out_file', 'in_file')]),
            (medial_nans, update_metadata, [('out_file', 'in_file')]),
        ])  # fmt:skip
    else:
        workflow.connect([(sampler, update_metadata, [('out_file', 'in_file')])])

    return workflow


def init_goodvoxels_bold_mask_wf(mem_gb: float, name: str = 'goodvoxels_bold_mask_wf'):
    """Calculate a mask of a BOLD series excluding high variance voxels.

    Workflow Graph
        .. workflow::
            :graph2use: colored
            :simple_form: yes

            from nibabies.workflows.bold.resampling import init_goodvoxels_bold_mask_wf
            wf = init_goodvoxels_bold_mask_wf(mem_gb=0.1)

    Parameters
    ----------
    mem_gb : :obj:`float`
        Size of BOLD file in GB
    name : :obj:`str`
        Name of workflow (default: ``goodvoxels_bold_mask_wf``)

    Inputs
    ------
    anat_ribbon
        Cortical ribbon in anatomical space
    bold_file
        Motion-corrected BOLD series in anatomical space

    Outputs
    -------
    masked_bold
        BOLD series after masking outlier voxels with locally high COV
    goodvoxels_ribbon
        Cortical ribbon mask excluding voxels with locally high COV
    """
    workflow = pe.Workflow(name=name)

    inputnode = pe.Node(
        niu.IdentityInterface(
            fields=[
                'anat_ribbon',
                'bold_file',
            ]
        ),
        name='inputnode',
    )
    outputnode = pe.Node(
        niu.IdentityInterface(
            fields=[
                'goodvoxels_mask',
                'goodvoxels_ribbon',
            ]
        ),
        name='outputnode',
    )
    ribbon_boldsrc_xfm = pe.Node(
        ApplyTransforms(interpolation='MultiLabel', transforms='identity'),
        name='ribbon_boldsrc_xfm',
        mem_gb=mem_gb,
    )

    stdev_volume = pe.Node(
        fsl.maths.StdImage(dimension='T'),
        name='stdev_volume',
        mem_gb=DEFAULT_MEMORY_MIN_GB,
    )

    mean_volume = pe.Node(
        fsl.maths.MeanImage(dimension='T'),
        name='mean_volume',
        mem_gb=DEFAULT_MEMORY_MIN_GB,
    )

    cov_volume = pe.Node(
        fsl.maths.BinaryMaths(operation='div'),
        name='cov_volume',
        mem_gb=DEFAULT_MEMORY_MIN_GB,
    )

    cov_ribbon = pe.Node(
        fsl.ApplyMask(),
        name='cov_ribbon',
        mem_gb=DEFAULT_MEMORY_MIN_GB,
    )

    cov_ribbon_mean = pe.Node(
        fsl.ImageStats(op_string='-M'),
        name='cov_ribbon_mean',
        mem_gb=DEFAULT_MEMORY_MIN_GB,
    )

    cov_ribbon_std = pe.Node(
        fsl.ImageStats(op_string='-S'),
        name='cov_ribbon_std',
        mem_gb=DEFAULT_MEMORY_MIN_GB,
    )

    cov_ribbon_norm = pe.Node(
        fsl.maths.BinaryMaths(operation='div'),
        name='cov_ribbon_norm',
        mem_gb=DEFAULT_MEMORY_MIN_GB,
    )

    smooth_norm = pe.Node(
        fsl.maths.MathsCommand(args='-bin -s 5'),
        name='smooth_norm',
        mem_gb=DEFAULT_MEMORY_MIN_GB,
    )

    merge_smooth_norm = pe.Node(
        niu.Merge(1),
        name='merge_smooth_norm',
        mem_gb=DEFAULT_MEMORY_MIN_GB,
        run_without_submitting=True,
    )

    cov_ribbon_norm_smooth = pe.Node(
        fsl.maths.MultiImageMaths(op_string='-s 5 -div %s -dilD'),
        name='cov_ribbon_norm_smooth',
        mem_gb=DEFAULT_MEMORY_MIN_GB,
    )

    cov_norm = pe.Node(
        fsl.maths.BinaryMaths(operation='div'),
        name='cov_norm',
        mem_gb=DEFAULT_MEMORY_MIN_GB,
    )

    cov_norm_modulate = pe.Node(
        fsl.maths.BinaryMaths(operation='div'),
        name='cov_norm_modulate',
        mem_gb=DEFAULT_MEMORY_MIN_GB,
    )

    cov_norm_modulate_ribbon = pe.Node(
        fsl.ApplyMask(),
        name='cov_norm_modulate_ribbon',
        mem_gb=DEFAULT_MEMORY_MIN_GB,
    )

    def _calc_upper_thr(in_stats):
        return in_stats[0] + (in_stats[1] * 0.5)

    upper_thr_val = pe.Node(
        Function(
            input_names=['in_stats'], output_names=['upper_thresh'], function=_calc_upper_thr
        ),
        name='upper_thr_val',
        mem_gb=DEFAULT_MEMORY_MIN_GB,
    )

    def _calc_lower_thr(in_stats):
        return in_stats[1] - (in_stats[0] * 0.5)

    lower_thr_val = pe.Node(
        Function(
            input_names=['in_stats'], output_names=['lower_thresh'], function=_calc_lower_thr
        ),
        name='lower_thr_val',
        mem_gb=DEFAULT_MEMORY_MIN_GB,
    )

    mod_ribbon_mean = pe.Node(
        fsl.ImageStats(op_string='-M'),
        name='mod_ribbon_mean',
        mem_gb=DEFAULT_MEMORY_MIN_GB,
    )

    mod_ribbon_std = pe.Node(
        fsl.ImageStats(op_string='-S'),
        name='mod_ribbon_std',
        mem_gb=DEFAULT_MEMORY_MIN_GB,
    )

    merge_mod_ribbon_stats = pe.Node(
        niu.Merge(2),
        name='merge_mod_ribbon_stats',
        mem_gb=DEFAULT_MEMORY_MIN_GB,
        run_without_submitting=True,
    )

    bin_mean_volume = pe.Node(
        fsl.maths.UnaryMaths(operation='bin'),
        name='bin_mean_volume',
        mem_gb=DEFAULT_MEMORY_MIN_GB,
    )

    merge_goodvoxels_operands = pe.Node(
        niu.Merge(2),
        name='merge_goodvoxels_operands',
        mem_gb=DEFAULT_MEMORY_MIN_GB,
        run_without_submitting=True,
    )

    goodvoxels_thr = pe.Node(
        fsl.maths.Threshold(),
        name='goodvoxels_thr',
        mem_gb=mem_gb,
    )

    goodvoxels_mask = pe.Node(
        fsl.maths.MultiImageMaths(op_string='-bin -sub %s -mul -1'),
        name='goodvoxels_mask',
        mem_gb=mem_gb,
    )

    # make HCP-style "goodvoxels" mask in anat space for filtering outlier voxels
    # in bold timeseries, based on modulated normalized covariance
    workflow.connect([
        (inputnode, ribbon_boldsrc_xfm, [('anat_ribbon', 'input_image')]),
        (inputnode, stdev_volume, [('bold_file', 'in_file')]),
        (inputnode, mean_volume, [('bold_file', 'in_file')]),
        (mean_volume, ribbon_boldsrc_xfm, [('out_file', 'reference_image')]),
        (stdev_volume, cov_volume, [('out_file', 'in_file')]),
        (mean_volume, cov_volume, [('out_file', 'operand_file')]),
        (cov_volume, cov_ribbon, [('out_file', 'in_file')]),
        (ribbon_boldsrc_xfm, cov_ribbon, [('output_image', 'mask_file')]),
        (cov_ribbon, cov_ribbon_mean, [('out_file', 'in_file')]),
        (cov_ribbon, cov_ribbon_std, [('out_file', 'in_file')]),
        (cov_ribbon, cov_ribbon_norm, [('out_file', 'in_file')]),
        (cov_ribbon_mean, cov_ribbon_norm, [('out_stat', 'operand_value')]),
        (cov_ribbon_norm, smooth_norm, [('out_file', 'in_file')]),
        (smooth_norm, merge_smooth_norm, [('out_file', 'in1')]),
        (cov_ribbon_norm, cov_ribbon_norm_smooth, [('out_file', 'in_file')]),
        (merge_smooth_norm, cov_ribbon_norm_smooth, [('out', 'operand_files')]),
        (cov_ribbon_mean, cov_norm, [('out_stat', 'operand_value')]),
        (cov_volume, cov_norm, [('out_file', 'in_file')]),
        (cov_norm, cov_norm_modulate, [('out_file', 'in_file')]),
        (cov_ribbon_norm_smooth, cov_norm_modulate, [('out_file', 'operand_file')]),
        (cov_norm_modulate, cov_norm_modulate_ribbon, [('out_file', 'in_file')]),
        (ribbon_boldsrc_xfm, cov_norm_modulate_ribbon, [('output_image', 'mask_file')]),
        (cov_norm_modulate_ribbon, mod_ribbon_mean, [('out_file', 'in_file')]),
        (cov_norm_modulate_ribbon, mod_ribbon_std, [('out_file', 'in_file')]),
        (mod_ribbon_mean, merge_mod_ribbon_stats, [('out_stat', 'in1')]),
        (mod_ribbon_std, merge_mod_ribbon_stats, [('out_stat', 'in2')]),
        (merge_mod_ribbon_stats, upper_thr_val, [('out', 'in_stats')]),
        (merge_mod_ribbon_stats, lower_thr_val, [('out', 'in_stats')]),
        (mean_volume, bin_mean_volume, [('out_file', 'in_file')]),
        (upper_thr_val, goodvoxels_thr, [('upper_thresh', 'thresh')]),
        (cov_norm_modulate, goodvoxels_thr, [('out_file', 'in_file')]),
        (bin_mean_volume, merge_goodvoxels_operands, [('out_file', 'in1')]),
        (goodvoxels_thr, goodvoxels_mask, [('out_file', 'in_file')]),
        (merge_goodvoxels_operands, goodvoxels_mask, [('out', 'operand_files')]),
    ])  # fmt:skip

    goodvoxels_ribbon_mask = pe.Node(
        fsl.ApplyMask(),
        name_source=['in_file'],
        keep_extension=True,
        name='goodvoxels_ribbon_mask',
        mem_gb=DEFAULT_MEMORY_MIN_GB,
    )

    # apply goodvoxels ribbon mask to bold
    workflow.connect([
        (goodvoxels_mask, goodvoxels_ribbon_mask, [('out_file', 'in_file')]),
        (ribbon_boldsrc_xfm, goodvoxels_ribbon_mask, [('output_image', 'mask_file')]),
        (goodvoxels_mask, outputnode, [('out_file', 'goodvoxels_mask')]),
        (goodvoxels_ribbon_mask, outputnode, [('out_file', 'goodvoxels_ribbon')]),
    ])  # fmt:skip

    return workflow


def init_bold_fsLR_resampling_wf(
    grayord_density: ty.Literal['91k', '170k'],
    omp_nthreads: int,
    mem_gb: float,
    name: str = 'bold_fsLR_resampling_wf',
):
    """Resample BOLD time series to fsLR surface.

    This workflow is derived heavily from three scripts within the DCAN-HCP pipelines scripts

    Line numbers correspond to the locations of the code in the original scripts, found at:
    https://github.com/DCAN-Labs/DCAN-HCP/tree/9291324/

    Workflow Graph
        .. workflow::
            :graph2use: colored
            :simple_form: yes

            from nibabies.workflows.bold.resampling import init_bold_fsLR_resampling_wf
            wf = init_bold_fsLR_resampling_wf(
                grayord_density='91k',
                omp_nthreads=1,
                mem_gb=1,
            )

    Parameters
    ----------
    grayord_density : :class:`str`
        Either ``"91k"`` or ``"170k"``, representing the total *grayordinates*.
    omp_nthreads : :class:`int`
        Maximum number of threads an individual process may use
    mem_gb : :class:`float`
        Size of BOLD file in GB
    name : :class:`str`
        Name of workflow (default: ``bold_fsLR_resampling_wf``)

    Inputs
    ------
    bold_file : :class:`str`
        Path to BOLD file resampled into anatomical space
    white : :class:`list` of :class:`str`
        Path to left and right hemisphere white matter GIFTI surfaces.
    pial : :class:`list` of :class:`str`
        Path to left and right hemisphere pial GIFTI surfaces.
    midthickness : :class:`list` of :class:`str`
        Path to left and right hemisphere midthickness GIFTI surfaces.
    midthickness_fsLR : :class:`list` of :class:`str`
        Path to left and right hemisphere midthickness GIFTI surfaces in fsLR space.
    sphere_reg_fsLR : :class:`list` of :class:`str`
        Path to left and right hemisphere sphere.reg GIFTI surfaces, mapping from subject to fsLR
    cortex_mask : :class:`list` of :class:`str`
        Path to left and right hemisphere cortical masks.
    volume_roi : :class:`str` or Undefined
        Pre-calculated goodvoxels mask. Not required.

    Outputs
    -------
    bold_fsLR : :class:`list` of :class:`str`
        Path to BOLD series resampled as functional GIFTI files in fsLR space

    """
    import smriprep
    import templateflow.api as tf
    from niworkflows.engine.workflows import LiterateWorkflow as Workflow
    from niworkflows.interfaces.utility import KeySelect
    from niworkflows.interfaces.workbench import VolumeToSurfaceMapping

    fslr_density = '32k' if grayord_density == '91k' else '59k'

    workflow = Workflow(name=name)

    workflow.__desc__ = """\
The BOLD time-series were resampled onto the left/right-symmetric template
"fsLR" using the Connectome Workbench [@hcppipelines].
"""

    inputnode = pe.Node(
        niu.IdentityInterface(
            fields=[
                'bold_file',
                'white',
                'pial',
                'midthickness',
                'midthickness_fsLR',
                'sphere_reg_fsLR',
                'cortex_mask',
                'volume_roi',
            ]
        ),
        name='inputnode',
    )

    hemisource = pe.Node(
        niu.IdentityInterface(fields=['hemi']),
        name='hemisource',
        iterables=[('hemi', ['L', 'R'])],
    )

    joinnode = pe.JoinNode(
        niu.IdentityInterface(fields=['bold_fsLR']),
        name='joinnode',
        joinsource='hemisource',
    )

    outputnode = pe.Node(
        niu.IdentityInterface(fields=['bold_fsLR']),
        name='outputnode',
    )

    # select white, midthickness and pial surfaces based on hemi
    select_surfaces = pe.Node(
        KeySelect(
            fields=[
                'white',
                'pial',
                'midthickness',
                'midthickness_fsLR',
                'sphere_reg_fsLR',
                'template_sphere',
                'cortex_mask',
                'template_roi',
            ],
            keys=['L', 'R'],
        ),
        name='select_surfaces',
        run_without_submitting=True,
    )
    select_surfaces.inputs.template_sphere = [
        str(sphere)
        for sphere in tf.get(
            template='fsLR',
            density=fslr_density,
            suffix='sphere',
            space=None,
            extension='.surf.gii',
        )
    ]
    atlases = smriprep.load_data('atlases')
    select_surfaces.inputs.template_roi = [
        str(atlases / f'L.atlasroi.{fslr_density}_fs_LR.shape.gii'),
        str(atlases / f'R.atlasroi.{fslr_density}_fs_LR.shape.gii'),
    ]

    # RibbonVolumeToSurfaceMapping.sh
    # Line 85 thru ...
    volume_to_surface = pe.Node(
        VolumeToSurfaceMapping(method='ribbon-constrained'),
        name='volume_to_surface',
        mem_gb=mem_gb * 3,
        n_procs=omp_nthreads,
    )
    metric_dilate = pe.Node(
        MetricDilate(distance=10, nearest=True),
        name='metric_dilate',
        mem_gb=1,
        n_procs=omp_nthreads,
    )
    mask_native = pe.Node(MetricMask(), name='mask_native')
    resample_to_fsLR = pe.Node(
        MetricResample(method='ADAP_BARY_AREA', area_surfs=True),
        name='resample_to_fsLR',
        mem_gb=1,
        n_procs=omp_nthreads,
    )
    # ... line 89
    mask_fsLR = pe.Node(MetricMask(), name='mask_fsLR')

    workflow.connect([
        (inputnode, select_surfaces, [
            ('white', 'white'),
            ('pial', 'pial'),
            ('midthickness', 'midthickness'),
            ('midthickness_fsLR', 'midthickness_fsLR'),
            ('sphere_reg_fsLR', 'sphere_reg_fsLR'),
            ('cortex_mask', 'cortex_mask'),
        ]),
        (hemisource, select_surfaces, [('hemi', 'key')]),
        # Resample BOLD to native surface, dilate and mask
        (inputnode, volume_to_surface, [
            ('bold_file', 'volume_file'),
            ('volume_roi', 'volume_roi'),
        ]),
        (select_surfaces, volume_to_surface, [
            ('midthickness', 'surface_file'),
            ('white', 'inner_surface'),
            ('pial', 'outer_surface'),
        ]),
        (select_surfaces, metric_dilate, [('midthickness', 'surf_file')]),
        (select_surfaces, mask_native, [('cortex_mask', 'mask')]),
        (volume_to_surface, metric_dilate, [('out_file', 'in_file')]),
        (metric_dilate, mask_native, [('out_file', 'in_file')]),
        # Resample BOLD to fsLR and mask
        (select_surfaces, resample_to_fsLR, [
            ('sphere_reg_fsLR', 'current_sphere'),
            ('template_sphere', 'new_sphere'),
            ('midthickness', 'current_area'),
            ('midthickness_fsLR', 'new_area'),
            ('cortex_mask', 'roi_metric'),
        ]),
        (mask_native, resample_to_fsLR, [('out_file', 'in_file')]),
        (select_surfaces, mask_fsLR, [('template_roi', 'mask')]),
        (resample_to_fsLR, mask_fsLR, [('out_file', 'in_file')]),
        # Output
        (mask_fsLR, joinnode, [('out_file', 'bold_fsLR')]),
        (joinnode, outputnode, [('bold_fsLR', 'bold_fsLR')]),
    ])  # fmt:skip

    return workflow


def init_bold_grayords_wf(
    grayord_density: ty.Literal['91k', '170k'],
    repetition_time: float,
    name: str = 'bold_grayords_wf',
):
    """
    Sample Grayordinates files onto the fsLR atlas.

    Outputs are in CIFTI2 format.

    Workflow Graph
        .. workflow::
            :graph2use: colored
            :simple_form: yes

            from nibabies.workflows.bold.resampling import init_bold_grayords_wf
            wf = init_bold_grayords_wf(grayord_density='91k')

    Parameters
    ----------
    grayord_density : :obj:`str`
        Either `91k` or `170k`, representing the total of vertices or *grayordinates*.
    mem_gb : :obj:`float`
        Size of BOLD file in GB
    name : :obj:`str`
        Unique name for the subworkflow (default: ``'bold_grayords_wf'``)

    Inputs
    ------
    bold_std : :obj:`str`
        The subcortical structures in MNI152NLin6Asym space.
    bold_labels : :obj:`str`
        Volume file containing all subcortical labels
    bold_fsLR : :obj:`list`
        List of BOLD files resampled on the fsaverage (ico7) surfaces

    Outputs
    -------
    dtseries : :obj:`str`
        BOLD CIFTI dense timeseries.
    dtseries_metadata : :obj:`str`
        BIDS metadata file corresponding to ``dtseries``.

    """
    import templateflow.api as tf
    from niworkflows.engine.workflows import LiterateWorkflow as Workflow

    from nibabies.interfaces.nibabel import ReorientImage
    from nibabies.interfaces.workbench import CiftiCreateDenseTimeseries

    workflow = Workflow(name=name)
    workflow.__desc__ = f"""\
*Grayordinates* files [@hcppipelines] containing {grayord_density} samples were also
generated using the highest-resolution ``fsaverage`` as intermediate standardized
surface space.
"""

    fslr_density = '32k' if grayord_density == '91k' else '59k'

    inputnode = pe.Node(
        niu.IdentityInterface(fields=['bold_std', 'bold_labels', 'bold_fsLR']),
        name='inputnode',
    )

    outputnode = pe.Node(
        niu.IdentityInterface(fields=['dtseries', 'dtseries_metadata']),
        name='outputnode',
    )

    split_surfaces = pe.Node(
        niu.Function(function=_split_surfaces, output_names=['left_surface', 'right_surface']),
        name='split_surfaces',
    )

    reorient_data = pe.Node(ReorientImage(target_orientation='LAS'), name='reorient_data')
    reorient_labels = reorient_data.clone(name='reorient_labels')

    gen_cifti = pe.Node(CiftiCreateDenseTimeseries(timestep=repetition_time), name='gen_cifti')
    gen_cifti.inputs.roi_left = tf.get(
        'fsLR',
        density=fslr_density,
        hemi='L',
        desc='nomedialwall',
        suffix='dparc',
    )
    gen_cifti.inputs.roi_right = tf.get(
        'fsLR',
        density=fslr_density,
        hemi='R',
        desc='nomedialwall',
        suffix='dparc',
    )
    gen_cifti_metadata = pe.Node(
        niu.Function(function=_gen_metadata, output_names=['out_metadata']),
        name='gen_cifti_metadata',
    )
    gen_cifti_metadata.inputs.grayord_density = grayord_density

    workflow.connect([
        (inputnode, reorient_data, [('bold_std', 'in_file')]),
        (inputnode, reorient_labels, [('bold_labels', 'in_file')]),
        (reorient_data, gen_cifti, [('out_file', 'volume_data')]),
        (reorient_labels, gen_cifti, [('out_file', 'volume_structure_labels')]),
        (inputnode, split_surfaces, [('bold_fsLR', 'in_surfaces')]),
        (split_surfaces, gen_cifti, [
            ('left_surface', 'left_metric'),
            ('right_surface', 'right_metric')]),
        (gen_cifti, outputnode, [('out_file', 'dtseries')]),
        (gen_cifti_metadata, outputnode, [('out_metadata', 'dtseries_metadata')]),
    ])  # fmt:skip
    return workflow


def _gen_metadata(grayord_density):
    import json
    from pathlib import Path

    from niworkflows.interfaces.cifti import _prepare_cifti

    _, _, metadata = _prepare_cifti(grayord_density)
    metadata_file = Path('bold.dtseries.json').absolute()
    metadata_file.write_text(json.dumps(metadata, indent=2))
    return str(metadata_file)


def _split_surfaces(in_surfaces: list) -> tuple[str, str]:
    """Split surfaces into selectable hemispheres."""
    return in_surfaces[0], in_surfaces[1]
